/*
*
*  * Copyright (c) 2019. Honeywell International, Inc. All Rights Reserved.
*
*/

package com.honeywell.easytemplating.components;

import com.honeywell.easytemplating.enums.BEasyTemplatingToastEnum;
import com.honeywell.easytemplating.px.BSimpleObjectFile;
import com.honeywell.easytemplating.util.BEasyTemplatingToast;
import com.honeywell.easytemplating.util.EasyTemplatingBuilderException;
import com.honeywell.easytemplating.util.EasyTemplatingGenerationHelper;
import com.honeywell.easytemplating.util.EasyTemplatingLogger;
import com.honeywell.easytemplating.util.EasyTemplatingUtils;

import javax.baja.file.BIFile;
import javax.baja.gx.BImage;
import javax.baja.naming.BOrd;
import javax.baja.nre.annotations.NiagaraAction;
import javax.baja.nre.annotations.NiagaraProperty;
import javax.baja.nre.annotations.NiagaraType;
import javax.baja.sys.Action;
import javax.baja.sys.BComponent;
import javax.baja.sys.BInteger;
import javax.baja.sys.BString;
import javax.baja.sys.Flags;
import javax.baja.sys.Property;
import javax.baja.sys.Sys;
import javax.baja.sys.Type;
import javax.baja.util.Lexicon;

import static com.honeywell.easytemplating.util.EasyTemplatingConst.*;

@NiagaraProperty(
        name = "templateName",
        type = "BString",
        defaultValue = "BString.make(\"\")",
        flags = Flags.SUMMARY
)

@NiagaraProperty(
        name = "templatePx",
        type = "BOrd",
        defaultValue = "BOrd.NULL",
        flags = Flags.SUMMARY
)

@NiagaraProperty(
        name = "popupPx",
        type = "BOrd",
        defaultValue = "BOrd.NULL",
        flags = Flags.SUMMARY
)

@NiagaraAction(
        name = "generateTemplate",
        flags = Flags.SUMMARY
)
@NiagaraType
public class BEasyTemplate extends BComponent {
    /*+ ------------ BEGIN BAJA AUTO GENERATED CODE ------------ +*/
    /*@ $com.honeywell.easytemplating.components.BEasyTemplate(3621327530)1.0$ @*/
    /* Generated Mon Apr 29 14:17:38 BST 2019 by Slot-o-Matic (c) Tridium, Inc. 2012 */

////////////////////////////////////////////////////////////////
// Property "templateName"
////////////////////////////////////////////////////////////////

    /**
     * Slot for the {@code templateName} property.
     *
     * @see #getTemplateName
     * @see #setTemplateName
     */
    public static final Property templateName = newProperty(Flags.SUMMARY, BString.make(""), null);

    /**
     * Get the {@code templateName} property.
     *
     * @see #templateName
     */
    public String getTemplateName() {
        return getString(templateName);
    }

    /**
     * Set the {@code templateName} property.
     *
     * @see #templateName
     */
    public void setTemplateName(String v) {
        setString(templateName, v, null);
    }

////////////////////////////////////////////////////////////////
// Property "templatePx"
////////////////////////////////////////////////////////////////

    /**
     * Slot for the {@code templatePx} property.
     *
     * @see #getTemplatePx
     * @see #setTemplatePx
     */
    public static final Property templatePx = newProperty(Flags.SUMMARY, BOrd.NULL, null);

    /**
     * Get the {@code templatePx} property.
     *
     * @see #templatePx
     */
    public BOrd getTemplatePx() {
        return (BOrd) get(templatePx);
    }

    /**
     * Set the {@code templatePx} property.
     *
     * @see #templatePx
     */
    public void setTemplatePx(BOrd v) {
        set(templatePx, v, null);
    }

////////////////////////////////////////////////////////////////
// Property "popupPx"
////////////////////////////////////////////////////////////////

    /**
     * Slot for the {@code popupPx} property.
     *
     * @see #getPopupPx
     * @see #setPopupPx
     */
    public static final Property popupPx = newProperty(Flags.SUMMARY, BOrd.NULL, null);

    /**
     * Get the {@code popupPx} property.
     *
     * @see #popupPx
     */
    public BOrd getPopupPx() {
        return (BOrd) get(popupPx);
    }

    /**
     * Set the {@code popupPx} property.
     *
     * @see #popupPx
     */
    public void setPopupPx(BOrd v) {
        set(popupPx, v, null);
    }

////////////////////////////////////////////////////////////////
// Action "generateTemplate"
////////////////////////////////////////////////////////////////

    /**
     * Slot for the {@code generateTemplate} action.
     *
     * @see #generateTemplate()
     */
    public static final Action generateTemplate = newAction(Flags.SUMMARY, null);

    /**
     * Invoke the {@code generateTemplate} action.
     *
     * @see #generateTemplate
     */
    public void generateTemplate() {
        invoke(generateTemplate, null, null);
    }

////////////////////////////////////////////////////////////////
// Type
////////////////////////////////////////////////////////////////

    @Override
    public Type getType() {
        return TYPE;
    }

    public static final Type TYPE = Sys.loadType(BEasyTemplate.class);

    /*+ ------------ END BAJA AUTO GENERATED CODE -------------- +*/

    private static final Lexicon lex = Lexicon.make(BEasyTemplate.class);
    /**
     * Generate template callback function
     */
    public void doGenerateTemplate() {

        BComponent root;
        try {

            BOrd templatePxOrd = getTemplatePx();
            BOrd popupPxOrd = getPopupPx();
            String provideTemplateName = getTemplateName();

            if (provideTemplateName.length() <= 0 || templatePxOrd.isNull()) {
                BEasyTemplatingToast.raise(
                        BString.make(lex.getText(ET_TOAST_ERR_TITLE)),
                        BImage.make(ET_TOAST_ICON_ORD_ERROR),
                        BString.make(lex.getText(ET_TEMPLATE_NAME_TEMPLATE_PX)),
                        BString.make(lex.getText(ET_FILL_FIELDS)),
                        BInteger.make(ET_TOAST_LONG_TIMEOUT),
                        BEasyTemplatingToastEnum.make(BEasyTemplatingToastEnum.ERROR));
                return;
            }

            // Get root of template
            root = getCommonRootComponent(templatePxOrd, popupPxOrd);

            if (Sys.getStation().asComponent().equals(root)) {
                BEasyTemplatingToast.raise(
                        BString.make(lex.getText(ET_TOAST_ERR_TITLE)),
                        BImage.make(ET_TOAST_ICON_ORD_ERROR),
                        BString.make(lex.getText(ET_TEMPLATE_CREATION_ERR)),
                        BString.make(lex.getText(ET_TEMPLATE_ROOT_ERR)),
                        BInteger.make(ET_TOAST_LONG_TIMEOUT),
                        BEasyTemplatingToastEnum.make(BEasyTemplatingToastEnum.ERROR));
                return;
            }

            // Generate neql
            String neql = EasyTemplatingUtils.getNeqlQueryFromComponentHierarchy(root);

            // Generate template
            EasyTemplatingGenerationHelper.generateTemplate(root, provideTemplateName, neql, templatePxOrd, popupPxOrd);
        } catch (Exception e) {
            EasyTemplatingLogger.infoWithException("Unknown exception occurred during Easy Template generation", e);
            BEasyTemplatingToast.raise(
                    BString.make(lex.getText(ET_TOAST_ERR_TITLE)),
                    BImage.make(ET_TOAST_ICON_ORD_ERROR),
                    BString.make(lex.getText(ET_CREATE_IN_MEMORY_TEMPLATE_FILE)),
                    BString.make(e.getMessage()),
                    BInteger.make(ET_TOAST_LONG_TIMEOUT),
                    BEasyTemplatingToastEnum.make(BEasyTemplatingToastEnum.ERROR));
        }

    }

    private static BComponent getCommonRootComponent(BOrd template, BOrd popup)
            throws EasyTemplatingBuilderException {
        BIFile templateBIFile = (BIFile) template.resolve().get();

        if (templateBIFile.getType().is(BSimpleObjectFile.TYPE)) {
            throw new IllegalArgumentException("Simple Object cannot be use to create Easy Template!");
        }

        if (!popup.isNull()) {

            BIFile popupBIFile = (BIFile) popup.resolve().get();

            return (BComponent) EasyTemplatingUtils.getCommonRootBComponent(false, false, templateBIFile,
                    popupBIFile).resolve().get();
        }

        return (BComponent) EasyTemplatingUtils.getCommonRootBComponent(false, false, templateBIFile).resolve().get();
    }

    public BEasyTemplate() {
        //fw required
    }
}
