/*
*
*  * Copyright (c) 2019. Honeywell International, Inc. All Rights Reserved.
*
*/
package com.honeywell.easytemplating.components;

import com.honeywell.easytemplating.util.EasyTemplatingBuilderException;
import com.honeywell.easytemplating.util.EasyTemplatingLogger;
import com.honeywell.easytemplating.util.EasyTemplatingUtils;
import com.tridium.template.BTemplateConfig;

import javax.baja.control.BControlPoint;
import javax.baja.naming.BOrd;
import javax.baja.control.BIWritablePoint;
import javax.baja.nre.annotations.NiagaraAction;
import javax.baja.nre.annotations.NiagaraProperty;
import javax.baja.nre.annotations.NiagaraType;
import javax.baja.sys.Type;
import javax.baja.sys.Flags;
import javax.baja.sys.BComponent;
import javax.baja.sys.Property;
import javax.baja.sys.BAbsTime;
import javax.baja.sys.Action;
import javax.baja.sys.Clock;
import javax.baja.sys.Sys;
import javax.baja.sys.BRelTime;
import javax.baja.sys.RelationKnob;
import javax.baja.sys.BLink;
import javax.baja.util.BFolder;
import java.util.HashMap;
import java.util.Map;

import static com.honeywell.easytemplating.util.EasyTemplatingConst.*;


@NiagaraProperty(name = "LastOkTime", type = "BAbsTime", defaultValue = "BAbsTime.DEFAULT", flags = 1)
@NiagaraProperty(name = "LastFailTime", type = "BAbsTime", defaultValue = "BAbsTime.DEFAULT", flags = 1)
@NiagaraProperty(name = "LastFailCause", type = "baja:String", defaultValue = "", flags = 1)
@NiagaraProperty(name = "isTemplateChanged", type = "boolean", defaultValue = "false", flags = Flags.HIDDEN)

@NiagaraAction(name = "LinkPoints")
@NiagaraAction(name = "ResetVariables",
        parameterType = "BComponent",
        defaultValue = "new BComponent()",
        flags = Flags.READONLY | Flags.HIDDEN)

@NiagaraType
public class BEasyTemplatingLinker extends BComponent {

    /*+ ------------ BEGIN BAJA AUTO GENERATED CODE ------------ +*/
    /*@ $com.honeywell.easytemplating.components.BEasyTemplatingLinker(2006652417)1.0$ @*/
    /* Generated Tue May 28 12:29:07 BST 2019 by Slot-o-Matic (c) Tridium, Inc. 2012 */

////////////////////////////////////////////////////////////////
// Property "LastOkTime"
////////////////////////////////////////////////////////////////

    /**
     * Slot for the {@code LastOkTime} property.
     *
     * @see #getLastOkTime
     * @see #setLastOkTime
     */
    public static final Property LastOkTime = newProperty(1, BAbsTime.DEFAULT, null);

    /**
     * Get the {@code LastOkTime} property.
     *
     * @see #LastOkTime
     */
    public BAbsTime getLastOkTime() {
        return (BAbsTime) get(LastOkTime);
    }

    /**
     * Set the {@code LastOkTime} property.
     *
     * @see #LastOkTime
     */
    public void setLastOkTime(BAbsTime v) {
        set(LastOkTime, v, null);
    }

////////////////////////////////////////////////////////////////
// Property "LastFailTime"
////////////////////////////////////////////////////////////////

    /**
     * Slot for the {@code LastFailTime} property.
     *
     * @see #getLastFailTime
     * @see #setLastFailTime
     */
    public static final Property LastFailTime = newProperty(1, BAbsTime.DEFAULT, null);

    /**
     * Get the {@code LastFailTime} property.
     *
     * @see #LastFailTime
     */
    public BAbsTime getLastFailTime() {
        return (BAbsTime) get(LastFailTime);
    }

    /**
     * Set the {@code LastFailTime} property.
     *
     * @see #LastFailTime
     */
    public void setLastFailTime(BAbsTime v) {
        set(LastFailTime, v, null);
    }

////////////////////////////////////////////////////////////////
// Property "LastFailCause"
////////////////////////////////////////////////////////////////

    /**
     * Slot for the {@code LastFailCause} property.
     *
     * @see #getLastFailCause
     * @see #setLastFailCause
     */
    public static final Property LastFailCause = newProperty(1, "", null);

    /**
     * Get the {@code LastFailCause} property.
     *
     * @see #LastFailCause
     */
    public String getLastFailCause() {
        return getString(LastFailCause);
    }

    /**
     * Set the {@code LastFailCause} property.
     *
     * @see #LastFailCause
     */
    public void setLastFailCause(String v) {
        setString(LastFailCause, v, null);
    }

////////////////////////////////////////////////////////////////
// Property "isTemplateChanged"
////////////////////////////////////////////////////////////////

    /**
     * Slot for the {@code isTemplateChanged} property.
     *
     * @see #getIsTemplateChanged
     * @see #setIsTemplateChanged
     */
    public static final Property isTemplateChanged = newProperty(Flags.HIDDEN, false, null);

    /**
     * Get the {@code isTemplateChanged} property.
     *
     * @see #isTemplateChanged
     */
    public boolean getIsTemplateChanged() {
        return getBoolean(isTemplateChanged);
    }

    /**
     * Set the {@code isTemplateChanged} property.
     *
     * @see #isTemplateChanged
     */
    public void setIsTemplateChanged(boolean v) {
        setBoolean(isTemplateChanged, v, null);
    }

////////////////////////////////////////////////////////////////
// Action "LinkPoints"
////////////////////////////////////////////////////////////////

    /**
     * Slot for the {@code LinkPoints} action.
     *
     * @see #LinkPoints()
     */
    public static final Action LinkPoints = newAction(0, null);

    /**
     * Invoke the {@code LinkPoints} action.
     *
     * @see #LinkPoints
     */
    public void LinkPoints() {
        invoke(LinkPoints, null, null);
    }

////////////////////////////////////////////////////////////////
// Action "ResetVariables"
////////////////////////////////////////////////////////////////

    /**
     * Slot for the {@code ResetVariables} action.
     *
     * @see #ResetVariables(BComponent parameter)
     */
    public static final Action ResetVariables = newAction(Flags.READONLY | Flags.HIDDEN, new BComponent(), null);

    /**
     * Invoke the {@code ResetVariables} action.
     *
     * @see #ResetVariables
     */
    public void ResetVariables(BComponent parameter) {
        invoke(ResetVariables, parameter, null);
    }

////////////////////////////////////////////////////////////////
// Type
////////////////////////////////////////////////////////////////

    @Override
    public Type getType() {
        return TYPE;
    }

    public static final Type TYPE = Sys.loadType(BEasyTemplatingLinker.class);

    /*+ ------------ END BAJA AUTO GENERATED CODE -------------- +*/


    private Clock.Ticket linkTicket = null;
    private Clock.Ticket variableTicket = null;

    /**
     * Helper Function: Checks if there are any links exist attached to the points in the provided template.
     */
    private boolean checkIfTheLinksExist(BComponent template) throws EasyTemplatingBuilderException {
        Map<String, BComponent> sourceChildren = getChildComponentRecursive(template);
        for (Map.Entry<String, BComponent> templatePoint : sourceChildren.entrySet()) {
            if (templatePoint.getValue().isRelationTarget()) {
                return true;
            }
        }
        return false;
    }

    /**
     * Public Function: Checks if the parent folder is a template and got links.
     * In case if there are no links set, attaches linkTicket.
     */
    @Override
    public void started() throws Exception {
        discardTicketIfActive();
        discardVariableTicketIfActive();

        BComponent parent = (BComponent) this.getParent();
        BComponent[] childComponents = parent.getChildComponents();
        Map<String, Integer> childComponentNames = new HashMap<>();
        for (int i = 0; i < childComponents.length; i++) {
            childComponentNames.put(childComponents[i].getNavName(), i);
        }
        if (childComponentNames.containsKey("templateConfig") && !checkIfTheLinksExist(parent)) {
            linkTicket = Clock.schedulePeriodically(this, BRelTime.makeSeconds(PARENT_CHECK_DELAY), LinkPoints, null);
        }

        if (!getIsTemplateChanged()) {
            variableTicket = Clock.schedulePeriodically(this, BRelTime.makeSeconds(1), ResetVariables, parent);
        }
    }

    /**
     * Public Function: Stops the ticket if component is removed.
     */
    @Override
    public void stopped() throws Exception {
        discardTicketIfActive();
        discardVariableTicketIfActive();
        super.stopped();
    }

    /**
     * Action call back looking for etso files and change the variables in it
     */
    public void doResetVariables(BComponent root) {

        EasyTemplatingUtils.resetComplexObjectEtsoBindings(root);

        //Stop process if variables are changed
        discardVariableTicketIfActive();
        //set true, make sure never will be re-write again
        setIsTemplateChanged(true);

        //base template folder re-writing when duplicated template deployment happened
        Property templateConfigProperty = root.getProperty("templateConfig");
        BTemplateConfig templateConfig = ((BTemplateConfig) root.get(templateConfigProperty));
        if (!root.getName().equalsIgnoreCase(templateConfig.getTemplateName())) {
            BFolder baseTemplateDir = (BFolder) BOrd.make(BASE_SLOT + templateConfig.getTemplateName())
                    .resolve().get();
            EasyTemplatingUtils.resetComplexObjectEtsoBindings(baseTemplateDir.asComponent());
        }
    }

    public void doLinkPoints() {
        for (RelationKnob relation : ((BComponent) this.getParent()).getRelationKnobs()) {
            if (ET_TAG_PARENT.equals(relation.getRelationId())) {
                try {
                    linkTemplatePoints(relation.getRelationComponent());
                    discardTicketIfActive();
                    return;
                } catch (Exception e) {
                    EasyTemplatingLogger.error(ET_ERROR_LINK_FAILED + e.getMessage(), e);
                    setLastFailTime(BAbsTime.now());
                    setLastFailCause(ET_ERROR_NORELATION);
                }
            }
        }

        setLastFailTime(BAbsTime.now());
        setLastFailCause(ET_ERROR_NORELATION);
    }

    private void discardTicketIfActive() {
        if (linkTicket != null) {
            linkTicket.cancel();
            linkTicket = null;
        }
    }

    private void discardVariableTicketIfActive() {
        if (variableTicket != null) {
            variableTicket.cancel();
            variableTicket = null;
        }
    }

    private void linkTemplatePoints(BComponent sourceRoot) throws EasyTemplatingBuilderException {
        Map<String, BComponent> templateChildren = getChildComponentRecursive((BComponent) getParent());
        Map<String, BComponent> sourceChildren = getChildComponentRecursive(sourceRoot);

        for (Map.Entry<String, BComponent> templateEntry : templateChildren.entrySet()) {
            BComponent templatePoint = templateEntry.getValue();
            BComponent sourcePoint = sourceChildren.get(templatePoint.getName());
            String linkId = "link" + Long.toString(BAbsTime.now().getMillis());

            if (templatePoint instanceof BIWritablePoint) {
                // Create link from source point's OUT slot to template point's IN2 slot
                BLink sourceToTemplateLink =
                        new BLink(sourcePoint.getSlotPathOrd(),
                                ET_SLOT_OUT,
                                ET_SLOT_IN2,
                                true);
                templateEntry.getValue().add(linkId, sourceToTemplateLink);
            }

            if (sourcePoint instanceof BIWritablePoint) {
                // Create link from template point's OUT slot to source point's IN16 slot
                BLink templateToSourceLink =
                        new BLink(templateEntry.getValue().getSlotPathOrd(),
                                ET_SLOT_OUT,
                                ET_SLOT_IN16,
                                true);
                sourcePoint.add(linkId, templateToSourceLink);
            }
        }

        setLastOkTime(BAbsTime.now());
    }

    /**
     * Helper function: recursively search the provided root component for child BControlPoints
     */
    private Map<String, BComponent> getChildComponentRecursive(BComponent root) throws EasyTemplatingBuilderException {
        Map<String, BComponent> returnComponents = new HashMap<>();

        BComponent[] childComponents = root.getChildComponents();
        for (BComponent child : childComponents) {
            if (child.getType().equals(BFolder.TYPE)) {
                returnComponents.putAll(getChildComponentRecursive(child));
            } else if (child.getType().is(BControlPoint.TYPE)) {
                returnComponents.put(child.getName(), child);
            }
        }
        return returnComponents;
    }

    private static final int PARENT_CHECK_DELAY = 15;
}


