/*
*
*  * Copyright (c) 2019. Honeywell International, Inc. All Rights Reserved.
*
*/
package com.honeywell.easytemplating.util;

import com.honeywell.easytemplating.enums.BEasyTemplatingToastEnum;

import javax.baja.gx.BBrush;
import javax.baja.gx.BColor;
import javax.baja.gx.BFont;
import javax.baja.gx.BImage;
import javax.baja.nre.annotations.AgentOn;
import javax.baja.nre.annotations.NiagaraType;
import javax.baja.sys.BInteger;
import javax.baja.sys.BString;
import javax.baja.sys.Context;
import javax.baja.sys.Sys;
import javax.baja.sys.Type;
import javax.baja.ui.BLabel;
import javax.baja.ui.BWidget;
import javax.baja.ui.enums.BHalign;
import javax.baja.ui.pane.BBorderPane;
import javax.baja.ui.pane.BGridPane;
import javax.baja.ui.pane.BPane;
import javax.baja.util.BNotification;
import javax.baja.workbench.BWbShell;
import javax.baja.workbench.util.BNotificationHandler;
import javax.baja.workbench.util.BNotifyPane;

import static com.honeywell.easytemplating.util.EasyTemplatingConst.*;

@NiagaraType(agent = @AgentOn(types = {"easyTemplating:EasyTemplatingToast"}))

public class BEasyTemplatingToastHandler extends BNotificationHandler {
    /*+ ------------ BEGIN BAJA AUTO GENERATED CODE ------------ +*/
    /*@ $com.honeywell.easytemplating.util.BEasyTemplatingToastHandler(3069913909)1.0$ @*/
    /* Generated Tue Apr 23 11:31:59 BST 2019 by Slot-o-Matic (c) Tridium, Inc. 2012 */

////////////////////////////////////////////////////////////////
// Type
////////////////////////////////////////////////////////////////

    @Override
    public Type getType() {
        return TYPE;
    }

    public static final Type TYPE = Sys.loadType(BEasyTemplatingToastHandler.class);

    /*+ ------------ END BAJA AUTO GENERATED CODE -------------- +*/

    private static BFont titleFont = BFont.make(BFont.make(BFont.DEFAULT, BFont.BOLD), ET_TOAST_TITLE_FONT_SIZE);
    private static BFont messageFont = BFont.make(BFont.DEFAULT, ET_TOAST_MESSAGE_FONT_SIZE);

    /**
     * Respond to a notification.
     */
    @Override
    public void handle(BWbShell shell, BNotification notify, Context cx) {
        BString title = (BString) notify.get(ET_TOAST_TITLE_KEY);
        BImage icon = (BImage) notify.get(ET_TOAST_ICON_KEY);
        BString message1 = (BString) notify.get(ET_TOAST_MESSAGE1_KEY);
        BString message2 = (BString) notify.get(ET_TOAST_MESSAGE2_KEY);
        BInteger timeoutMs = (BInteger) notify.get(ET_TOAST_TIMEOUT_MS_KEY);
        BEasyTemplatingToastEnum type = (BEasyTemplatingToastEnum) notify.get(ET_TOAST_TYPE_KEY);

        String nTitle = (title != null) ? title.getString() : ET_TITLE;
        BImage nIcon = (icon != null) ? icon : BImage.make(ET_TOAST_ICON_ORD_INFO);
        String nMessage1 = (message1 != null) ? message1.getString() : "";
        String nMessage2 = (message2 != null) ? message2.getString() : "";
        int nTimeoutMs = (timeoutMs != null) ? timeoutMs.getInt() : ET_TOAST_SHORT_TIMEOUT;

        BNotifyPane pane = new BNotifyPane();
        pane.setTop(makeTitleWidget(nTitle, type));
        pane.setCenter(makeMessageWidget(nIcon, nMessage1, nMessage2));
        pane.open(nTimeoutMs);
    }

    private BWidget makeTitleWidget(String title, BEasyTemplatingToastEnum type) {
        BLabel label = new BLabel(title);
        label.setHalign(BHalign.center);
        label.setBackground(BBrush.makeSolid(getHeaderColor(type)));
        label.setForeground(BBrush.makeSolid(ET_TOAST_TITLE_FG_COLOR));
        label.setPadding(ET_TOAST_TITLE_INSETS);
        label.setFont(titleFont);
        return label;
    }

    private BColor getHeaderColor(BEasyTemplatingToastEnum type) {
        if (ET_TOAST_SUCCESS == type.getOrdinal()) {
            return ET_TOAST_TITLE_SUCCESS_BG_COLOR;
        } else if (ET_TOAST_WARNING == type.getOrdinal()) {
            return ET_TOAST_TITLE_WARNING_BG_COLOR;
        }

        return ET_TOAST_TITLE_ERROR_BG_COLOR;
    }

    private BWidget makeMessageWidget(BImage icon, String msg1, String msg2) {
        BGridPane grid = new BGridPane(1);
        grid.setColumnAlign(BHalign.center);
        grid.add(null, new BLabel(icon));
        maybeAddMessage(grid, msg1);
        maybeAddMessage(grid, msg2);
        BBorderPane pane = new BBorderPane(grid);
        pane.setPadding(ET_TOAST_MESSAGE_WIDGET_INSETS);
        return pane;
    }

    private void maybeAddMessage(BPane pane, String msg) {
        if (msg.length() > 0) {
            BLabel label = new BLabel(msg);
            label.setFont(messageFont);
            label.setHalign(BHalign.center);
            pane.add(null, label);
        }
    }

}
