/*
 *
 *  * Copyright (c) 2019. Honeywell International, Inc. All Rights Reserved.
 *
 */
package com.honeywell.easytemplating.util;

import com.honeywell.easytemplating.components.BEasyTemplatingLinker;
import com.honeywell.easytemplating.px.BSimpleObjectFile;
import com.tridium.fox.sys.file.BFoxFileSpace;
import com.tridium.kitpx.BPopupBinding;
import com.tridium.sys.registry.NAgentList;
import com.tridium.template.BTemplateConfig;
import com.tridium.template.file.PxFileRef;

import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import javax.baja.agent.AgentInfo;
import javax.baja.agent.BPxView;
import javax.baja.control.BControlPoint;
import javax.baja.file.BDirectory;
import javax.baja.file.BFileSystem;
import javax.baja.file.BIFile;
import javax.baja.file.FilePath;
import javax.baja.file.types.text.BPxFile;
import javax.baja.naming.BOrd;
import javax.baja.naming.OrdQuery;
import javax.baja.naming.SlotPath;
import javax.baja.nav.BINavNode;
import javax.baja.space.BSpace;
import javax.baja.sys.BComponent;
import javax.baja.sys.BFacets;
import javax.baja.sys.BStation;
import javax.baja.sys.BValue;
import javax.baja.sys.BasicContext;
import javax.baja.sys.Property;
import javax.baja.sys.Slot;
import javax.baja.sys.Sys;
import javax.baja.sys.Type;
import javax.baja.ui.BBinding;
import javax.baja.ui.BValueBinding;
import javax.baja.ui.BWidget;
import javax.baja.ui.px.BPxInclude;
import javax.baja.ui.px.PxDecoder;
import javax.baja.ui.px.PxEncoder;
import javax.baja.util.BFolder;
import javax.baja.util.Lexicon;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.channels.FileChannel;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.TreeMap;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.HashSet;

import static com.honeywell.easytemplating.util.EasyTemplatingConst.BASE_DEPLOY;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.BASE_SLOT;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_BINDING_ORD_PREFIX;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_CHUNK_TWO;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_COMPLEX_OBJ_FILE_EXT;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_DEPLOY_DIRECTORY_ORD;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_EASY_ALARM_BINDING_TYPE;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_EASY_BINDING_WIDGET_TYPE;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_EASY_HISTORY_POPUP_BINDING;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_EASY_OVERRIDE_BINDING_TYPE;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_EASY_POPUP_BINDING_TYPE;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_EASY_TEMPLATING_BINDING_TYPE;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_EASY_VALUE_BINDING_TYPE;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_ETSO_BINDING_STRING;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_FILE;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_FILE_DEPLOY_DIRECTORY_ORD;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_FILE_OS;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_HISTORY_OS;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_HON_BACNET_HISTORY_BINDING;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_HON_BACNET_VIRTUAL_PROPERTY_BINDING;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_KITPX_POPUP_BINDING_NAME;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_LINKER;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_NEQL_AND;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_NEQL_CHARS_QUOTE;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_NEQL_CHECK_EMPTY_FOLDERS;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_NEQL_CHECK_NAME;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_NEQL_CHECK_TYPE;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_ORD;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_POPUP_BINDING;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_SIMPLE_OBJ_FILE_EXT;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_VARIABLES;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_VIEW_REQ_EXT;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_WB_COMPONENT_SPACE;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.ET_WB_LOCAL_SPACE;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.WB_FILE_BASE_SLOT;

import static com.honeywell.easytemplating.util.EasyTemplatingConst.EB_EASYBINDINGWIDGET;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.BOUNDLABEL;
import static com.honeywell.easytemplating.util.EasyTemplatingConst.VALUEBINDING;
import static javax.xml.parsers.DocumentBuilderFactory.newInstance;

public final class EasyTemplatingUtils {

    private static int closingBracketsToAdd = 0;
    private static final Logger LOGGER = Logger.getLogger("EasyTemplate");
    private static final String ET_WB_LOCAL_SCHEME = "local:|";
    public static final String ET_WB_FOXS_SCHEME = "foxs:|";
    public static final String ET_WB_LOCAL_FOXS = ET_WB_LOCAL_SCHEME + ET_WB_FOXS_SCHEME;
    public static final Lexicon lex = Lexicon.make(EasyTemplatingUtils.class);
    private EasyTemplatingUtils() {

    }

    /*****************************************************************
     *  Public API
     *****************************************************************/
    /**
     * Change binding in file
     */
    public static void changeBIFileBindingAtDeployment(BIFile file, String templateName) {

        try {
            //get writable file
            BIFile wFile = EasyTemplatingUtils.getWritableBIFile(file);
            PxDecoder decodedPx = new PxDecoder(wFile);
            BWidget rootWidget = decodedPx.decodeDocument();

            List<BBinding> bindings = getBindingsRecursive(rootWidget);

            for (BBinding binding : bindings) {
                if (binding.getOrd().toString().contains("file")
                        && !binding.getOrd().toString().contains(templateName)) {
                    //replace "file:^px/deploy/<THIS>/fan2Popup.px"/
                    String removeBase = binding.getOrd().toString().replace(ET_FILE_DEPLOY_DIRECTORY_ORD, "");
                    String restOftheOrd = removeBase.substring(removeBase.indexOf('/'), removeBase.length());
                    StringBuilder newOrd = new StringBuilder();
                    newOrd.append(ET_FILE_DEPLOY_DIRECTORY_ORD)
                            .append(templateName.toLowerCase(Locale.ENGLISH)).append(restOftheOrd);
                    binding.set(ET_ORD, BOrd.make(newOrd.toString()));
                } else if (deepCheck(binding, templateName)) {
                    String[] chunks = binding.getOrd().toString().split("/");

                    StringBuilder newOrd = new StringBuilder();
                    //replace root with template name in ord
                    replaceRootNameInOrd(newOrd, chunks, templateName, 1);

                    binding.set(ET_ORD, BOrd.make(newOrd.toString()));
                }
            }
            //close file
            decodedPx.close();

            PxEncoder encodedPx = new PxEncoder(wFile.getOutputStream());
            encodedPx.encodeDocument(rootWidget);
            encodedPx.close();

            if(!EasyTemplatingUtils.getFile(wFile).setReadOnly()){
                EasyTemplatingLogger.info("Unable to set file readonly: "+wFile.getFilePath());
            }
        } catch (Exception e) {
            EasyTemplatingLogger.infoWithException("Change file binding failed", e);
        }
    }

    /**
     * Re-set Easy Template deployed based on Complex object
     * change variable to point to the root component
     */
    public static void resetComplexObjectEtsoBindings(BComponent root) {

        //if no etso file in the template or first time deployment (no folder exists) do not process !!!
        if (!isEtsoFileUsedOrFirstDeployyment(root)) {
            return;
        }
        //Check and create a new directory in px/deploy with copied files if need
        checkDeployedFolderExists(root);
        BPxView view = null;
        NAgentList agents = (NAgentList) root.getAgents(new BasicContext());

        for (AgentInfo info : agents) {
            //Check fo PxInclude
            if (info.getAgentType().is(BPxView.TYPE)) {
                view = (BPxView) info;
                //Stop for loop if px view found
                break;
            }
        }

        try {
            String baseView = getBasePxView(view, root);

            StringBuilder newOrd = new StringBuilder();
            String[] chunks = baseView.split("/");
            replaceRootNameInOrd(newOrd, chunks, root.getName(), ET_CHUNK_TWO);
            //writable BIFile
            BIFile wFile = EasyTemplatingUtils.getWritableBIFile(((BIFile) BOrd.make(newOrd.toString())
                    .get(Sys.getStation())));
            PxDecoder decodedPx = new PxDecoder(wFile);
            BWidget rootWidget = decodedPx.decodeDocument();

            //Function used to remove complexity (SonarLint)
            resetBindings(rootWidget, root, decodedPx, newOrd.toString());

        } catch (Exception e) {
            EasyTemplatingLogger.infoWithException("Scheduler etso variable changer, " +
                    "Can't resolve the file: retry in 1sec", e);
        }
    }

    /**
     * Collect all PXInclude files from BWidget -> OK
     * add sub folders popup.px files as well
     */
    public static List<BWidget> getAllPxIncludedFilesWidget(BWidget root) {

        List<BWidget> returnOrds = new ArrayList<>();

        BWidget[] childWidgets = root.getChildWidgets();
        for (BWidget child : childWidgets) {
            if (child.getChildWidgets().length > 0) {
                returnOrds.addAll(getAllPxIncludedFilesWidget(child));
            }
            returnOrds.addAll(getPxIncludeWidgets(Arrays.asList(child.getChildWidgets())));
        }

        return returnOrds;
    }

    /**
     * Check root folder for existing linker
     */
    public static void addLinkerToComponent(BComponent comp) {
        boolean existingLinker = false;
        //Check for linker in the destination BFolder
        BComponent[] compChildComponents = comp.getChildComponents();
        for (BComponent component : compChildComponents) {
            if (component.getType().is(BEasyTemplatingLinker.TYPE)) {
                existingLinker = true;
            }
        }
        if (!existingLinker) {
            comp.add(ET_LINKER, new BEasyTemplatingLinker());
        }
    }

    /**
     * This function adding the popup binding to the Object PX Pop
     *
     * @param root
     * @param objectPxOrd
     * @throws Exception
     */
    public static void addPopupBindingToObject(String root, BOrd objectPxOrd,
                                               String popupBindingPrefix)
            throws EasyTemplatingBuilderException {
        try {
            //writable BIFile
            BIFile wFile = EasyTemplatingUtils.getWritableBIFile((BIFile) objectPxOrd.get(Sys.getStation()));
            PxDecoder decodedPx = new PxDecoder(wFile);

            BWidget rootWidget = decodedPx.decodeDocument();

            //root level child binding
            for (BComponent child : rootWidget.getChildComponents()) {
                //child binding
                if (containsBinding(Arrays.asList(child.getChildComponents()), ET_POPUP_BINDING)) {
                    addBinding(root, child, popupBindingPrefix);
                }
                //children of the child bindings
                addPopupBindingToChildComponent(root, child, popupBindingPrefix);
            }

            decodedPx.close();

            PxEncoder encodedPx = new PxEncoder(wFile.getOutputStream());
            encodedPx.encodeDocument(rootWidget);

            encodedPx.close();

            if(!EasyTemplatingUtils.getFile(wFile).setReadOnly()){
                EasyTemplatingLogger.info("Unable to set file readonly: "+wFile.getFilePath());
            }
        } catch (Exception ex) {
            throw new EasyTemplatingBuilderException(ex.getMessage(), ex);
        }
    }

    /**
     * Public function: generate final NEQL query
     */
    public static String getNeqlQueryFromComponentHierarchy(BComponent rootComponent) {
        StringBuilder resultingNeqlQuery = new StringBuilder();

        List<List<BComponent>> componentGroupsBasedOnDepth = getComponentGroupsBasedOnDepth(rootComponent, 1);

        for (int currentLevel = 0; currentLevel < componentGroupsBasedOnDepth.size(); currentLevel++) {
            List<BComponent> currentLevelPoints = componentGroupsBasedOnDepth.get(currentLevel);
            int pointsPerLevel = currentLevelPoints.size();
            StringBuilder pointNeqlQuery = new StringBuilder();

            if (pointsPerLevel == 0) {
                pointNeqlQuery.append(getLevelBasedNeqlQueryWithNoPoints(currentLevel));
            } else {
                pointNeqlQuery.append(getLevelBasedNeqlQueryWithPoints(currentLevel, currentLevelPoints));
            }

            resultingNeqlQuery.append(pointNeqlQuery);
            if (currentLevel < componentGroupsBasedOnDepth.size() - 1) {
                resultingNeqlQuery.append(ET_NEQL_AND);
            }
        }

        return resultingNeqlQuery.toString();
    }

    /**
     * Public function: assigns dynamic PX file as a view to the root component
     */
    public static void addDynamicViewToTheComponent(BComponent rootComponent, BOrd pxFileOrd, String viewName) {

        if (!containsSlot(Arrays.asList(rootComponent.getSlotsArray()), viewName)) {
            BPxView dynamicPxView = new BPxView();
            dynamicPxView.setPxFile(pxFileOrd);

            rootComponent.add(viewName, dynamicPxView);
        }
    }


    public static boolean isRelativised(String ordString) {
        return ordString.startsWith("slot:");
    }
    //***********************************************************************************************
    // Finding the Root Component
    //***********************************************************************************************

    /**
     * Return Common BComponent from root PX file
     */
    public static BOrd getCommonRootBComponent(boolean isSimpleObjectCreation,
                                               boolean isWorkbench, BIFile... selectedFile)
            throws EasyTemplatingBuilderException {

        // Init global variables

        pxFiles.clear();
        popupPxFiles.clear();

        // Get all bindings from all px files
        // 1) Not empty
        // 2) No null ords
        // 3) Contains absolute ords:
        //     a) No BEasyTemplatingBinding
        //     b) No BPxInclude
        //     c) Not only BControlPoint, but also the BComponent that used to show PLV
        //     d) BControlPoint's property already been parsed to BControlPoint's ord
        //     e) No BEasyValueBinding whose parent widget is not BEasyBindingWidget
        //     f) No BEasyAlarmBinding whose parent widget is not BEasyBindingWidget
        //     g) No BEasyOverrideBinding whose parent widget is not BEasyBindingWidget
        //     h) Support HonBacnetVirtualPropertyBinding
        //     i) Support HonBacnetHistoryBinding
        // 4) Contains relative ords:
        //     a) Contains all ords which start with "slot:"
        //     b) May be "slot:", "slot:NumericWritable", "slot:NumericWritable/set"
        //     c) Do not make sure ord is BControlPoint

        List<BOrd> allBindings = getAllBindingsFromObjectAndPopupFiles(true, isWorkbench, selectedFile);

        if (allBindings.isEmpty()) {
            return BOrd.NULL;
        }

        // Classify ords into absolute and relative

        List<BOrd> absoluteOrds = new ArrayList<>();
        List<BOrd> relativeOrds = new ArrayList<>();
        classifyOrdsIntoAbsoluteAndRelative(allBindings, absoluteOrds, relativeOrds);

        // Calculate common root for absolute ords
        // And it will be used as the common root of relative ords

        if (absoluteOrds.isEmpty() && !relativeOrds.isEmpty()) {
            return BOrd.NULL;
        }

        // Now absoluteOrds will be never empty

        allBindings = absoluteOrds;
        BOrd currentParent = allBindings.get(0);

            for(int i=0;i<allBindings.size();i++){
                currentParent = allBindings.get(i);
                if(!currentParent.toString().contains(ET_HISTORY_OS)) {
                    break;
                }
            }
        if (isControlPoint(currentParent)) {
            currentParent = currentParent.getParent();
        }

        while (currentParent != null) {
            if (!bindingsAllContainParent(allBindings, currentParent)) {
                currentParent = currentParent.getParent();
            } else {
                break;
            }
        }

        if (currentParent != null) {
            if (currentParent.toString().equals(BASE_SLOT)) {
                return BOrd.make(BASE_SLOT);
            } else {
                //add all files from root -> subFolders (views and popup.px files)
                addRootSubFoldersViewsAndPopupFiles(isSimpleObjectCreation, currentParent);
                return currentParent;
            }
        } else {
            return BOrd.make(BASE_SLOT);
        }
    }

    private static void classifyOrdsIntoAbsoluteAndRelative(List<BOrd> inputOrds, List<BOrd> absoluteOrds, List<BOrd> relativeOrds) {
        for (BOrd ord : inputOrds) {
            if (isRelativised(ord.toString())) {
                relativeOrds.add(ord);
            } else {
                absoluteOrds.add(ord);
            }
        }
    }

    /**
     * Recursively searches for bindings at all levels below root widget
     *
     * @param root The root level of the PxPage
     * @return List of BBindings used BWidget and children of BWidget
     * @throws Exception
     */
    public static List<BBinding> getBindingsRecursive(BWidget root) throws EasyTemplatingBuilderException {
        List<BBinding> returnBindings = new ArrayList<>();

        BWidget[] childWidgets = root.getChildWidgets();
        for (BWidget child : childWidgets) {
            if (child.getChildWidgets().length > 0) {
                returnBindings.addAll(getBindingsRecursive(child));
            }
            returnBindings.addAll(Arrays.asList(child.getBindings()));
        }

        return returnBindings;
    }

    /**
     * Collect all bindings
     */
    public static List<BOrd> getEtsoBindingsRecursive(boolean isWorkbench, BWidget root, BIFile main)
            throws EasyTemplatingBuilderException {

        List<BOrd> returnOrds = new ArrayList<>();

        BWidget[] childWidgets = root.getChildWidgets();
        for (BWidget child : childWidgets) {
            if (child.getChildWidgets().length > 0) {
                returnOrds.addAll(getEtsoBindingsRecursive(isWorkbench, child, main));
            }

            returnOrds.addAll(getEtsoFilesBindings(Arrays.asList(child.getChildWidgets()), main, isWorkbench));
        }


        return returnOrds;
    }

    /**
     * Collect all PXInclude files from BWidget
     */
    public static List<BOrd> getAllPxIncludedFilesOrd(BWidget root) {

        List<BOrd> returnOrds = new ArrayList<>();

        BWidget[] childWidgets = root.getChildWidgets();
        for (BWidget child : childWidgets) {
            if (child.getChildWidgets().length > 0) {
                returnOrds.addAll(getAllPxIncludedFilesOrd(child));
            }

            returnOrds.addAll(getPxIncludeFiles(Arrays.asList(child.getChildWidgets())));
        }

        return returnOrds;
    }

    public static List<BBinding> removePopupBinding(BWidget root) throws EasyTemplatingBuilderException {
        List<BBinding> returnBindings = new ArrayList<>();
        BWidget[] childWidgets = root.getChildWidgets();
        for (BWidget childWidget : childWidgets) {
            for(BWidget widget: childWidget.getChildWidgets()){
                for(Property widgetProperty: widget.getProperties()){
                    if(widgetProperty.getName().equals(ET_KITPX_POPUP_BINDING_NAME)){
                        widget.remove(widgetProperty);
                    }
                }
            }
            removePopupBinding(childWidget);
        }

        return returnBindings;
    }

    /**
     * Exclude all BEasyValueBinding whose parent widget is not BEasyBindingWidget
     * Exclude all BEasyAlarmBinding whose parent widget is not BEasyBindingWidget
     * Exclude all BEasyOverrideBinding whose parent widget is not BEasyBindingWidget
     */
    public static void excludeNeedlessEasyBindingInstance(List<BBinding> bindings) {
        bindings.removeIf(EasyTemplatingUtils::isNeedlessBindingInstanceOfEasyBindingWidget);
    }
    /**
     * Exclude all History popup bindings
     */
    public static void filterHistoryPopupBinding(List<BBinding> bindings) {
        bindings.removeIf(EasyTemplatingUtils::isHistoryPopupBinding);
    }

    /*****************************************************************
     *  Private API
     *****************************************************************/

    /**
     * Filter out BeasyTemplatingBinding type from the collection
     */
    private static List<BBinding> filterEasyTemplatingBindings(List<BBinding> bindings) {

        List<BBinding> returnBindings = new ArrayList<>();
        Type easyTemplatingBindingType = Sys.getType(ET_EASY_TEMPLATING_BINDING_TYPE);
        for (BBinding binding : bindings) {
            if (!binding.getType().is(easyTemplatingBindingType)) {
                returnBindings.add(binding);
            }
        }

        return returnBindings;
    }

    /**
     * Add all files from root -> sub folders views and popup.px files to the template!!!
     */
    private static void addRootSubFoldersViewsAndPopupFiles(boolean isSimpleObjectCreation,
                                                            BOrd rootOrd) throws EasyTemplatingBuilderException {

        if (isSimpleObjectCreation) {
            return;
        }

        BComponent root = (BComponent) rootOrd.resolve().get();
        BComponent[] childComponents = root.getChildComponents();

        for (BComponent child : childComponents) {

            if (child.getType().is(BFolder.TYPE)) {
                BPxView view = (BPxView) child.get("BPxView");
                if (view != null) {
                    addPxFiles(view.getPxFile());

                    BIFile viewFile = (BIFile) view.getPxFile().resolve().get();
                    getAllBindingOrdsFromPxFile(false, viewFile);
                }
            }
        }

    }

    /**
     * Get all included Px files Bord as List -> OK
     */
    private static List<BWidget> getPxIncludeWidgets(List<BWidget> widgets) {
        List<BWidget> allPxInclude = new ArrayList<>();
        for (BWidget widget : widgets) {
            if (widget.getType().is(BPxInclude.TYPE)) {
                allPxInclude.add(widget);
            }
        }
        return allPxInclude;
    }

    private static void changePxIncludeEntryVariable(BComponent root, BWidget pxIncludeEntry) throws IOException {

        Property variableProperty = pxIncludeEntry.getProperty(ET_VARIABLES);
        Property ordProperty = pxIncludeEntry.getProperty(ET_ORD);

        BComponent[] childes = root.getChildComponents();
        List<BComponent> directories = new ArrayList<>();
        directories.add(root);
        //collect all directories under root
        for (BComponent child : childes) {
            if (child.getType().is(BFolder.TYPE)) {
                directories.add(child);
            }
        }

        boolean isVariableChanged = false;

        String rootOrdString = "station:|" + root.getSlotPathOrd().toString();
        //directories contains all possible binding options
        for (BComponent value : directories) {

            if (isSameNameAsTarget(value.getName(), pxIncludeEntry.asComplex().get(ET_VARIABLES).toString())) {
                String valueOrdString = "station:|" + value.getSlotPathOrd().toString();
                changePxIncludeVariable(root, variableProperty, valueOrdString, pxIncludeEntry, ordProperty);
                isVariableChanged = true;
            }
        }

        if (!isVariableChanged) {
            changePxIncludeVariable(root, variableProperty, rootOrdString, pxIncludeEntry, ordProperty);
        }
    }

    private static void changePxIncludeVariable(BComponent root, Property variableProperty, String newVariable,
                                                BValue pxIncludeEntry, Property ordProperty) throws IOException {


        String ordPropValue = pxIncludeEntry.asComponent().get(ET_ORD).toString().replace(getTargetTemplateName(root)
                .toLowerCase(Locale.ENGLISH), root.getName().toLowerCase(Locale.ENGLISH));
        String rootVariable = pxIncludeEntry.asComponent().get(ET_VARIABLES).toString();
        String variableNewRoot = rootVariable.substring(0, rootVariable.indexOf('='));
        //new variable structure
        String decodedString = variableNewRoot + "=s:" + SlotPath.escape(newVariable);
        //set pxInclude variables to new value
        pxIncludeEntry.asComponent().set(variableProperty, BFacets.make(decodedString));
        pxIncludeEntry.asComponent().set(ordProperty, BOrd.make(ordPropValue));
    }

    private static boolean isSameNameAsTarget(String root, String entry) {
        return entry.contains(root);
    }

    /**
     * Helper method to resetComplexObjectEtsoBindings to remove complexity
     */
    private static boolean resetBindings(BWidget rootWidget, BComponent root, PxDecoder decodedPx, String newComplex) {
        try {
            //get the PxIncludes looking for variables and change the values
            List<BWidget> includedEtsoFiles = getAllPxIncludedFilesWidget(rootWidget);

            //if no PxInclude then return true
            if (includedEtsoFiles.isEmpty()) {
                return true;
            }

            changePxIncludeEntryVariable(includedEtsoFiles, root);

            decodedPx.close();

            BIFile wFile = EasyTemplatingUtils
                    .getWritableBIFile(((BIFile) BOrd.make(newComplex).get(Sys.getStation())));
            PxEncoder encodedPx = new PxEncoder(wFile.getOutputStream());
            encodedPx.encodeDocument(rootWidget);

            encodedPx.close();
            if(!EasyTemplatingUtils.getFile(wFile).setWritable(true,false)){
                EasyTemplatingLogger.info("Unable to set file writable: "+wFile.getFilePath());
            }

            //remove view
            root.remove(getBComponentView(root));
            //add view
            addDynamicViewToTheComponent(root, BOrd.make(newComplex), "MainView");

            //Get all sub folders
            List<BComponent> subFolders = getAllSubFoldersRecursive(root);
            //change views on subfolder if  any
            changeSubDiretoriesView(subFolders, root);
            //if process finished return true to close scheduler action
            return true;
        } catch (Exception e) {
            EasyTemplatingLogger.infoWithException("", e);
            return false;
        }
    }

    /**
     * Collect all BComponents within new deployed root
     */
    private static List<BComponent> getAllSubFoldersRecursive(BComponent root) {
        List<BComponent> subFolders = new ArrayList<>();

        BComponent[] childs = root.getChildComponents();
        for (BComponent child : childs) {
            if (child.getChildComponents().length > 0 && isContainsBFolder(child.getChildComponents())) {
                subFolders.addAll(getAllSubFoldersRecursive(child));
            }
            if (child.getType().is(BFolder.TYPE)) {
                subFolders.add(child);
            }

        }
        return subFolders;
    }

    private static boolean isContainsBFolder(BComponent[] childs) {
        return Arrays.stream(childs).anyMatch(comp -> comp.getType().is(BFolder.TYPE));
    }

    /**
     * loop through all root subdirectories and change the view on it to new deployed directory
     */
    private static void changeSubDiretoriesView(List<BComponent> subDirList, BComponent templateRoot) {

        for (BComponent child : subDirList) {
            if (child.getType().is(BFolder.TYPE)) {
                NAgentList agents = (NAgentList) child.getAgents(new BasicContext());
                for (AgentInfo info : agents) {
                    //Check fo PxInclude
                    changeBfolderView(info, templateRoot, child);
                }
            }
        }
    }

    /**
     * Check agent type for BPxView type
     */
    private static void changeBfolderView(AgentInfo info, BComponent templateRoot, BComponent child) {

        if (info.getAgentType().is(BPxView.TYPE)) {

            BPxView view = (BPxView) info;
            String newComplex = getBasePxView(view, templateRoot);

            //remove view
            child.remove(getBComponentView(child));
            //add view
            addDynamicViewToTheComponent(child, BOrd.make(newComplex), "MainView");
        }
    }


    /**
     * Change variable to point to the deployed folder
     */
    private static void changePxIncludeEntryVariable(List<BWidget> includedEtsoFiles, BComponent root)
            throws IOException {
        for (BWidget widget : includedEtsoFiles) {
            //change only etso files variable
            if (widget.get(ET_ORD).toString().contains(".etso")) {
                changePxIncludeEntryVariable(root, widget);
            }
        }
    }

    /**
     * Get BComponent view
     */
    private static Property getBComponentView(BComponent root) {
        Property[] props = root.getPropertiesArray();
        Property view = null;
        for (Property prop : props) {
            if (prop.getType().is(BPxView.TYPE)) {
                view = prop;
            }
        }
        return view;
    }

    /**
     * Returning the base template name from config
     */
    private static String getTargetTemplateName(BComponent root) {
        BComponent[] childs = root.getChildComponents();
        String tempName = "";
        for (BComponent comp : childs) {
            if (comp.getType().is(BTemplateConfig.TYPE)) {
                BTemplateConfig conf = (BTemplateConfig) comp;
                tempName = conf.getTemplateName();
                break;
            }
        }
        return tempName;
    }

    /**
     * Create a new directory to the newly deployed template
     */
    private static void makeNewTemplateDeployedDirectory(BComponent root) {

        try {
            //create a new directory with a new deploy name
            File newDir = new File(Sys.getStationHome() + BASE_DEPLOY + root.getName().toLowerCase(Locale.ENGLISH));
            //create a dir
            newDir.mkdir();
            //get the original template source directory
            File targetTemplateDir = new File(Sys.getStationHome() + BASE_DEPLOY + getTargetTemplateName(root));
            //get all files from the original directory
            File[] files = targetTemplateDir.listFiles();
            //copy all files from targetTempDir to new deployed directory
            for (File file : files) {
                String filename = String.format("%s%s%s" + File.separator + "%s", Sys.getStationHome(),
                        BASE_DEPLOY, root.getName(), file.getName());
                File destinationFile = new File(filename);

                //create a new empty file
                if (!destinationFile.createNewFile()) {
                    continue;
                }
                //Copy al files over to the new directory
                copyTemplateFiles(file, destinationFile);

                BIFile newDeployedFile = (BIFile) BOrd.make(ET_DEPLOY_DIRECTORY_ORD + "/" + root.getName()
                        + "/" + file.getName()).resolve().get();
                PxDecoder decodedPx = new PxDecoder(newDeployedFile);
                BWidget rootWidget = decodedPx.decodeDocument();

                if (!getBindingsRecursive(rootWidget).isEmpty()) {
                    changeBIFileBindingAtDeployment(newDeployedFile, root.getName());
                }
                decodedPx.close();
            }

        } catch (Exception e) {
            EasyTemplatingLogger.infoWithException("Can't create new directory", e);
        }
    }

    /**
     * Change to root like ["station:|slot:/", <TemplateName>, "sub1", "fan1"]
     */
    private static void replaceRootNameInOrd(StringBuilder newOrd, String[] chunks, String templateName, int position) {
        //["station:|slot:/", <TemplateName>, "sub1", "fan1"]
        chunks[position] = templateName.toLowerCase(Locale.ENGLISH);

        for (int i = 0; i < chunks.length; ++i) {
            newOrd.append(chunks[i]).append("/");
        }
        //delete trailing "/" slash at the end
        newOrd.deleteCharAt(newOrd.lastIndexOf("/"));
    }

    /**
     * return boolean -> not etso binding and no root in the binding and not relative to root
     */
    private static boolean deepCheck(BBinding binding, String templateName) {
        return !binding.getOrd().toString().contains(templateName)
                && !binding.getOrd().toString().contains(ET_ETSO_BINDING_STRING)
                && !binding.getOrd().toString().contains("..");
    }

    /**
     * Copy all files from - to
     */
    private static void copyTemplateFiles(File source, File destination) {

        try (FileChannel sourceChannel = new FileInputStream(source).getChannel();
             FileChannel destChannel = new FileOutputStream(destination).getChannel()) {

            destChannel.transferFrom(sourceChannel, 0, sourceChannel.size());

        } catch (IOException error) {
            EasyTemplatingLogger.error("File copy error", error);
        }

    }

    /**
     * Check for etso files
     */
    private static boolean isEtsoFileUsedOrFirstDeployyment(BComponent root) {

        boolean containsEtsoFile = false;
        try {

            BOrd.make(ET_DEPLOY_DIRECTORY_ORD + "/" + getTargetTemplateName(root)).resolve().get();
            //get the original template source directory
            File targetTemplateDir = new File(Sys.getStationHome() + BASE_DEPLOY + getTargetTemplateName(root));
            //get all files from the original directory
            File[] files = targetTemplateDir.listFiles();
            containsEtsoFile = containsEtsoFile(files);
        } catch (Exception e) {
            EasyTemplatingLogger.infoWithException("Folder not exists", e);
        }
        return containsEtsoFile;
    }

    /**
     * check files for etso file
     */
    private static boolean containsEtsoFile(File[] files) {
        return Arrays.stream(files).anyMatch(file -> file.getName().contains(ET_SIMPLE_OBJ_FILE_EXT));
    }

    /**
     * Try to get the newly deployed template corresponding folder
     * if no directory then create it
     */
    private static void checkDeployedFolderExists(BComponent root) {

        try {
            BOrd.make(ET_DEPLOY_DIRECTORY_ORD + "/" + root.getName()).resolve().get();
        } catch (Exception e) {
            EasyTemplatingLogger.infoWithException("Folder not exists", e);
            //create a new BFolder then return it back
            makeNewTemplateDeployedDirectory(root);
        }
    }

    /**
     * Return String: changed template base name with deployed name Ord as String
     */
    private static String getBasePxView(BPxView view, BComponent root) {
        String pxView = "";
        pxView = view.getPxFile().getOrd().toString().replace(getTargetTemplateName(root).toLowerCase(Locale.ENGLISH),
                root.getName().toLowerCase(Locale.ENGLISH));

        return pxView;
    }

    /**
     * Get all included Px files Bord as List
     */
    private static List<BOrd> getPxIncludeFiles(List<BWidget> widgets) {
        List<BOrd> allPxInclude = new ArrayList<>();
        for (BWidget widget : widgets) {
            if (widget.getType().is(BPxInclude.TYPE)) {
                    allPxInclude.add(BOrd.make(widget.get(ET_ORD).toString()));
            }
        }
        return allPxInclude;
    }
    /**
     * Collect all PXInclude files from BWidget
     */
    public static List<BOrd> getAllPxIncludedFilesOrdExcludingEtso(BWidget root) {

        List<BOrd> returnOrds = new ArrayList<>();

        BWidget[] childWidgets = root.getChildWidgets();
        for (BWidget child : childWidgets) {
            if (child.getChildWidgets().length > 0) {
                returnOrds.addAll(getAllPxIncludedFilesOrdExcludingEtso(child));
            }

            returnOrds.addAll(getPxIncludeFilesExcludingEtso(Arrays.asList(child.getChildWidgets())));
        }

        return returnOrds;
    }

    /**
     * Get all included Px files Bord as List
     */
    private static List<BOrd> getPxIncludeFilesExcludingEtso(List<BWidget> widgets) {
        List<BOrd> allPxInclude = new ArrayList<>();
        for (BWidget widget : widgets) {
            if (widget.getType().is(BPxInclude.TYPE)) {
                if(!widget.get(ET_ORD).toString().endsWith(".etso")) {
                    allPxInclude.add(BOrd.make(widget.get(ET_ORD).toString()));
                }
            }
        }
        return allPxInclude;
    }

    /**
     * PxInclude files get all BBindings, stop if the file not .etso file
     */
    private static List<BOrd> getEtsoFilesBindings(List<BWidget> widgets, BIFile main, boolean isWorkbench)
            throws EasyTemplatingBuilderException {

        List<BOrd> allBindings = new ArrayList<>();

        for (BWidget widget : widgets) {
            if (widget.getType().is(BPxInclude.TYPE)) {
                StringBuilder ordToResolve = new StringBuilder();
                ordToResolve.append(widget.get(ET_ORD).toString());
                if (isWorkbench) {
                    ordToResolve.setLength(0);
                    ordToResolve.append(ET_WB_COMPONENT_SPACE).append(widget.get(ET_ORD).toString());
                }

                BIFile etsoFile = (BIFile) BOrd.make(ordToResolve.toString()).resolve().get();

                /**Catch NOT .etso PxInclude files
                 * Extract information from PxInclude etso files
                 * */
                if (!etsoFile.getType().is(BSimpleObjectFile.TYPE)) {

                    throw new IllegalArgumentException("PxInclude " + etsoFile.getFileName() +
                            " not .etso file");
                }

                String variables = widget.get("variables").toString();
                if (!variables.contains(BASE_SLOT)) {
                    throw new IllegalArgumentException(main.getFileName() +
                            " No variables set to etso file(s)");
                }
                String var = variables.substring(variables.indexOf('=') + 1, variables.length());

                //PxInclude .etso file extraction
                try {
                    PxDecoder decodedPx = new PxDecoder(etsoFile);
                    BWidget rootWidget = decodedPx.decodeDocument();

                    allBindings.addAll(extractRootFromVarBinding(getBindingsRecursive(rootWidget), var));
                    decodedPx.close();
                } catch (Exception ex) {
                    throw new EasyTemplatingBuilderException(ex.getMessage(), ex);
                }
            }
        }
        return allBindings;
    }

    /**
     * Find PxInclude .etso files variables root
     */
    private static List<BOrd> extractRootFromVarBinding(List<BBinding> bindings, String var)
            throws EasyTemplatingBuilderException {

        List<BOrd> etsoOrds = new ArrayList<>();

        for (BBinding binding : bindings) {
            if (binding.get(ET_ORD).toString().contains(ET_ETSO_BINDING_STRING)) {
                String varOrd = binding.get(ET_ORD).toString();
                int closeVarTag = varOrd.indexOf(')');
                String restOfTheBOrd = varOrd.substring(closeVarTag + 1, varOrd.length());
                StringBuilder newBord = new StringBuilder();
                newBord.append(var).append(restOfTheBOrd);
                addBOrdToList(etsoOrds, BOrd.make(newBord.toString()));
            } else {
                BIFile file = (BIFile) binding.getOrd().resolve().get();
                if (!getAllBindingOrdsFromPxFile(false, file).isEmpty() && binding.getType().is(BPopupBinding.TYPE)) {
                    PxFileRef ref = new PxFileRef((BPxFile) file, binding.getOrd(), file.getFileName());
                    addPopupPxFiles(ref);
                } else {
                    //add etso file popup bindings to deploy file/px list
                    addPxFiles(binding.getOrd());
                }
            }
        }
        return etsoOrds;
    }

    /**
     * Check and add BOrd to a List<BOrd> if not containing
     */
    private static void addBOrdToList(List<BOrd> ords, BOrd ord) {
        if (!ords.contains(ord)) {
            ords.add(ord);
        }
    }

    /**
     * Helper function: recursively search the provided root component for child BControlPoints
     */
    private static Map<String, BComponent> getChildComponentWithDepthRecursive(BComponent root, int currentDepth) {
        Map<String, BComponent> returnComponents = new HashMap<>();
        BComponent[] childComponents = root.getChildComponents();
        for (BComponent child : childComponents) {
            if (child.getType().is(BControlPoint.TYPE)) {
                returnComponents.put(child.getName() + ',' + currentDepth, child);
            } else {
                returnComponents.putAll(getChildComponentWithDepthRecursive(child, currentDepth + 1));
            }
        }
        return returnComponents;
    }

    /**
     * Helper function: returns collection of same-level components in a sorted manner
     */
    private static List<List<BComponent>> getComponentGroupsBasedOnDepth(BComponent root, int depth) {
        Map<String, BComponent> childComponentEntries = getChildComponentWithDepthRecursive(root, depth);
        Map<Integer, List<BComponent>> componentsSortedByDepth = new TreeMap<>();
        List<List<BComponent>> componentGroupsBasedOnDepth = new ArrayList<>(new ArrayList<>());

        for (Map.Entry<String, BComponent> componentEntry : childComponentEntries.entrySet()) {
            String[] componentNameAndDepth = componentEntry.getKey().split(",");
            int currentDepth = Integer.parseInt(componentNameAndDepth[1]);
            BComponent currentComponent = componentEntry.getValue();
            if (componentsSortedByDepth.containsKey(currentDepth)) {
                List<BComponent> currentListAtThisDepth = componentsSortedByDepth.get(currentDepth);
                currentListAtThisDepth.add(currentComponent);
                componentsSortedByDepth.put(currentDepth, currentListAtThisDepth);
            } else {
                List<BComponent> currentListAtThisDepth = new ArrayList<>();
                currentListAtThisDepth.add(currentComponent);
                componentsSortedByDepth.put(currentDepth, currentListAtThisDepth);
            }
        }

        int minimumDepth = 0;
        Optional<Integer> optionalInteger = componentsSortedByDepth.keySet().stream().min(Integer::compareTo);
        if (optionalInteger.isPresent()) {
            minimumDepth = optionalInteger.get();
        }

        if (minimumDepth > 1) {
            while (componentGroupsBasedOnDepth.size() != minimumDepth - 1) {
                componentGroupsBasedOnDepth.add(new ArrayList<BComponent>());
            }
        }

        for (Map.Entry<Integer, List<BComponent>> componentWithDepth : componentsSortedByDepth.entrySet()) {
            componentGroupsBasedOnDepth.add(componentWithDepth.getValue());
        }

        return componentGroupsBasedOnDepth;
    }

    /**
     * Build popup binding String
     */
    private static BOrd buildPopupBindingString(String root, BOrd popupPxOrd, String popupBindingPrefix) {

        StringBuilder changedPopupPxFilePath = new StringBuilder();
        changedPopupPxFilePath.append(popupBindingPrefix + root);
        changedPopupPxFilePath.append(popupPxOrd.toString().substring(popupPxOrd.toString().lastIndexOf('/'),
                popupPxOrd.toString().length()));

        return BOrd.make(changedPopupPxFilePath.toString());
    }

    /**
     * Change bindings in child components
     */
    private static void addPopupBindingToChildComponent(String root, BComponent child,
                                                        String popupBindingPrefix) {
        for (BComponent innerChild : child.getChildComponents()) {
            if (containsBinding(Arrays.asList(innerChild.getChildComponents()), ET_POPUP_BINDING)) {
                addBinding(root, innerChild, popupBindingPrefix);
            }
        }
    }

    /**
     * Change binding in component level
     */
    private static void addBinding(String root, BComponent innerChild, String popupBindingPrefix) {
        BPopupBinding editedPopupBinding = new BPopupBinding();

        // get the actual XML node tag binding
        BPopupBinding originalPopupBinding = (BPopupBinding) innerChild.get(ET_POPUP_BINDING);
        // Change binding to relative only to the user selected Popup.px file
        BOrd popupOrdToRelativise = (BOrd) originalPopupBinding.get(ET_ORD);

        editedPopupBinding.setOrd(buildPopupBindingString(root, popupOrdToRelativise, popupBindingPrefix));

        innerChild.set(ET_POPUP_BINDING, editedPopupBinding);
    }


    private static boolean containsBinding(final List<BComponent> list, String propertyName) {
        return list.stream().anyMatch(o -> propertyName.equals(o.getName()));
    }

    private static boolean containsSlot(final List<Slot> list, String slotName) {
        return list.stream().anyMatch(o -> slotName.equals(o.getName()));
    }

    /**
     * Helper function: build a logically positive parent->child hierarchy
     */
    private static String buildPositiveLevelHierarchy(int currentLevel) {
        StringBuilder levelHierarchy = new StringBuilder();

        for (int levelCounter = 0; levelCounter < currentLevel + 1; levelCounter++) {
            levelHierarchy.append("n:child->");
        }

        return levelHierarchy.toString();
    }

    /**
     * Helper function: build a logically negative parent->child hierarchy
     */
    private static String buildNegativeLevelHierarchy(int currentLevel) {
        StringBuilder levelHierarchy = new StringBuilder();

        for (int levelCounter = 0; levelCounter < currentLevel + 1; levelCounter++) {
            if (levelCounter == currentLevel) {
                levelHierarchy.append("!(n:child->(");
            } else {
                levelHierarchy.append("(n:child->(");
            }
            closingBracketsToAdd++;
            closingBracketsToAdd++;
        }

        return levelHierarchy.toString();
    }

    /**
     * Helper function: append remaining closing brackets to the NEQL query
     */
    private static String appendRemainingClosingBrackets() {
        StringBuilder closingBrackets = new StringBuilder();

        while (closingBracketsToAdd != 0) {
            closingBrackets.append(")");
            closingBracketsToAdd--;
        }

        return closingBrackets.toString();
    }

    /**
     * Helper function: generate NEQL query with no points on current level
     */
    private static StringBuilder getLevelBasedNeqlQueryWithNoPoints(int currentLevel) {
        StringBuilder levelBasedNeqlQuery = new StringBuilder();

        levelBasedNeqlQuery.append(buildNegativeLevelHierarchy(currentLevel));
        levelBasedNeqlQuery.append("(n:point) or " + ET_NEQL_CHECK_EMPTY_FOLDERS);
        levelBasedNeqlQuery.append(appendRemainingClosingBrackets());

        return levelBasedNeqlQuery;
    }

    /**
     * Helper function: generate NEQL query given that there is at least 1 point on current level
     */
    private static StringBuilder getLevelBasedNeqlQueryWithPoints(int currentLevel,
                                                                  List<BComponent> currentLevelPoints) {
        StringBuilder levelBasedNeqlQuery = new StringBuilder();

        for (int i = 0; i < currentLevelPoints.size(); i++) {
            String currentPointName = String.format("%s%s%s", ET_NEQL_CHARS_QUOTE, currentLevelPoints.get(i).getName(),
                    ET_NEQL_CHARS_QUOTE);
            String currentPointType = String.format("%s%s%s", ET_NEQL_CHARS_QUOTE, currentLevelPoints.get(i).getType(),
                    ET_NEQL_CHARS_QUOTE);

            levelBasedNeqlQuery.append(buildPositiveLevelHierarchy(currentLevel));
            levelBasedNeqlQuery.append(ET_NEQL_CHECK_NAME)
                    .append(currentPointName)
                    .append(ET_NEQL_CHECK_TYPE)
                    .append(currentPointType)
                    .append(")");

            if (i < currentLevelPoints.size() - 1) {
                levelBasedNeqlQuery.append(ET_NEQL_AND);
            }
        }

        return levelBasedNeqlQuery;
    }

    /**
     * Returns whether all of the binding ORDS provided contain the common parent
     *
     * @param allBindings   all of the bindings which need to meet check
     * @param currentParent the common parent to check
     * @return boolean value
     */
    private static boolean bindingsAllContainParent(List<BOrd> allBindings, BOrd currentParent) {
        for (BOrd ord : allBindings) {
            if(ord.toString().contains(ET_HISTORY_OS)){
                continue;
            }
            boolean foundParent = false;
            BOrd currentOrd = ord;

            while (!foundParent) {
                if (currentOrd == null) {
                    return false;
                }

                if (currentOrd.equals(currentParent)) {
                    foundParent = true;
                } else {
                    currentOrd = currentOrd.getParent();
                }
            }
        }

        return true;
    }

    /**
     * Returns all bindings from both the Object and Popup PxFile
     *
     * @return List of BOrds from both provided files
     * @throws Exception
     */
    private static List<BOrd> getAllBindingsFromObjectAndPopupFiles(boolean isEtso, boolean isWorkbench, BIFile[] iFiles)
            throws EasyTemplatingBuilderException {

        List<BOrd> allPxBindings = new ArrayList<>();

        for (BIFile iFile : iFiles) {
            if (iFile == null) {
                continue;
            }

            if (isEtso) {
                allPxBindings.addAll(getAllBindingOrdsFromPxFile(isWorkbench, iFile));
            } else {
                allPxBindings.addAll(getAllBindingOrdsFromPxFileForComplex(isWorkbench, iFile));
            }

            if (allPxBindings.isEmpty()) {
                throw new IllegalArgumentException("No common root found");
            }
        }

        if (allPxBindings.isEmpty()) {
            throw new IllegalArgumentException("No common root found");
        }

        return allPxBindings;
    }

    /**
     * Gets all bindings used in a single PxFile
     *
     * @param file Ord of PxFile
     * @return List of BOrds used as bindings in PxFile
     * @throws Exception
     */
    private static List<BOrd> getAllBindingOrdsFromPxFile(boolean isWorkbench, BIFile file)
            throws EasyTemplatingBuilderException {
        List<BBinding> allBindings;

        try {
            PxDecoder decodedPx = new PxDecoder(file);
            BWidget rootWidget = decodedPx.decodeDocument();

            // Remove easy templating bindings
            allBindings = filterEasyTemplatingBindings(getBindingsRecursive(rootWidget));

            // Remove all BEasyValueBinding whose parent widget is not BEasyBindingWidget
            // Remove all BEasyAlarmBinding whose parent widget is not BEasyBindingWidget
            // Remove all BEasyOverrideBinding whose parent widget is not BEasyBindingWidget
            excludeNeedlessEasyBindingInstance(allBindings);

            // Get only control point ords, view popup ords (except file view ords), and all relative ords
            List<BOrd> allBord = getStationOrdsFromBindings(allBindings);

            // Always empty, only used to check PxInclude
            allBord.addAll(getBindingsOrdList(isWorkbench, rootWidget, file));

            return allBord;
        } catch (Exception ex) {
            throw new EasyTemplatingBuilderException(ex.getMessage(), ex);
        }
    }

    private static List<BOrd> getAllBindingOrdsFromPxFileForComplex(boolean isWorkbench, BIFile file)
        throws EasyTemplatingBuilderException {
        List<BBinding> allBindings;

        try {
            PxDecoder decodedPx = new PxDecoder(file);
            BWidget rootWidget = decodedPx.decodeDocument();

            // Get all bindings
            allBindings = getBindingsRecursive(rootWidget);

            // Remove all BEasyValueBinding whose parent widget is not BEasyBindingWidget
            // Remove all BEasyAlarmBinding whose parent widget is not BEasyBindingWidget
            // Remove all BEasyOverrideBinding whose parent widget is not BEasyBindingWidget
            excludeNeedlessEasyBindingInstance(allBindings);

            // Get only control point ords, view popup ords, and all relative ords
            List<BOrd> allBord = getStationOrdsFromBindings(allBindings);

            // Get EasyTemplatingBinding ords
            List<BOrd> easyTemplatingOrds = getAllEtsoBindings(file);
            allBord.addAll(easyTemplatingOrds);

            return allBord;
        } catch (Exception ex) {
            throw new EasyTemplatingBuilderException(ex.getMessage(), ex);
        }
    }

    /**
     * Return all bindings from station and check for PxInclude at SimpleObject creation 1st step
     */
    private static List<BOrd> getBindingsOrdList(boolean isWorkbench, BWidget rootWidget, BIFile file)
            throws EasyTemplatingBuilderException {
        if (isWorkbench && !getAllPxIncludedFilesOrd(rootWidget).isEmpty()) {
            throw new IllegalArgumentException("Simple Object cannot include Px file");
        }

        return getEtsoBindingsRecursive(isWorkbench, rootWidget, file);
    }

    /**
     * Gets only station ords from list of bindings
     *
     * @param bindings List of bindings of all types
     * @return Only bindings from within the station space
     */
    private static List<BOrd> getStationOrdsFromBindings(List<BBinding> bindings) {

        List<BOrd> returnOrds = new ArrayList<>();
        for (BBinding binding : bindings) {
            if (!isValidBinding(binding)) {
                continue;
            }
            // If ord is relative, do nothing but add to result list
            // Because there is no way to check a relative ord's real instance and type
            Map<String, BOrd> ordInfo = EasyTemplatingUtils.getRealOrd(binding);
            for (Map.Entry<String, BOrd> entrySet : ordInfo.entrySet()) {
                String slotName = entrySet.getKey();
                BOrd bindingOrd = entrySet.getValue();
                if (slotName.equals("hyperlink") && bindingOrd.isNull()) {
                    continue;
                }
                if (isRelativised(bindingOrd.toString())) {
                    returnOrds.add(bindingOrd);
                } else if (ET_FILE.equals(bindingOrd.parse()[0].getScheme())) {
                    addPxFiles(bindingOrd);
                } else if (bindingOrd.toString().contains(ET_VIEW_REQ_EXT)) {
                    String bindingOrdStr = bindingOrd.toString();
                    returnOrds.add(BOrd.make(WB_FILE_BASE_SLOT + bindingOrdStr.substring(0, bindingOrdStr.indexOf(ET_VIEW_REQ_EXT))));
                } else {
                    // Filter out the control point binding ords
                    BOrd newReturnOrd = getPointOrdForParentCalculation(binding, slotName);
                    if (newReturnOrd != BOrd.NULL) {
                        returnOrds.add(newReturnOrd);
                    }
                }
            }
        }
        return returnOrds;
    }

    private static boolean isValidBinding(BBinding binding) {
        Map<String, BOrd> ordInfo = EasyTemplatingUtils.getRealOrd(binding);
        for (Map.Entry<String, BOrd> entrySet : ordInfo.entrySet()) {
            String slotName = entrySet.getKey();
            BOrd ord = entrySet.getValue();
            if (slotName.equals("hyperlink") && ord.isNull()) {
                continue;
            }
            if (ord.equals(BOrd.NULL)) {
                return false;
            }
            try {
                if (!isRelativised(ord.toString())) {
                    BOrd.make(ET_WB_COMPONENT_SPACE + ord.toString()).resolve().get();
                }
            } catch (Exception ex) {
                if (LOGGER.isLoggable(Level.FINER)) {
                    LOGGER.log(Level.FINER, "Cannot resolve binding ord: " + ord.toString(), ex);
                }
                return false;
            }
        }
        return true;
    }

    private static BOrd getPointOrdForParentCalculation(BBinding binding, String slotName) {

        BOrd originalBindingOrd = (BOrd) binding.get(slotName);
        BOrd returnOrd = BOrd.make(WB_FILE_BASE_SLOT + originalBindingOrd.toString());

        while (!(returnOrd.get(Sys.getStation()) instanceof BControlPoint)) {
            returnOrd = returnOrd.getParent();

            if (returnOrd == null) {
                return BOrd.NULL;
            }
        }

        return returnOrd;
    }

    private static void addPxFiles(BOrd ord) {
        if (!pxFiles.contains(ord)) {
            pxFiles.add(ord);
        }
    }

    private static void addPopupPxFiles(PxFileRef ref) {
        if (!popupPxFiles.contains(ref)) {
            popupPxFiles.add(ref);
        }
    }

    public static boolean isSimpleObjectFile(BIFile file) {
        return ((null != file) && (file.getType().is(BSimpleObjectFile.TYPE)));
    }

    public static List<BOrd> getPxFiles() {
        return new ArrayList<>(pxFiles);
    }

    public static List<PxFileRef> getPopupPxFiles() {
        return new ArrayList<>(popupPxFiles);
    }

    /**
     * This function is called when the ET service started
     * it will set all etso files under Station etsoFiles directory (file:^etsoFiles) to readOnly()
     * and users cannot change it then through the WB editor
     */
    public static void setEtsoFilesToReadOnly(final File folder) {

        File[] files = folder.listFiles();

        for (File file : files) {
            if (file.isDirectory()) {
                setEtsoFilesToReadOnly(file);
            } else if (file.canWrite()) {
                if(!file.setReadOnly()){
                    EasyTemplatingLogger.info("Unable to set file readonly: "+file.getPath());
                }
            }
        }
    }

    /**
     * Pass a folder to the function to set all files under it to writable
     */
    public static void setEtsoFilesToWritable(final File folder) {

        File[] files = folder.listFiles();

        for (File file : files) {
            if (file.isDirectory()) {
                setEtsoFilesToReadOnly(file);
            } else if (file.canWrite()) {
                if(!file.setWritable(true, false)){
                    EasyTemplatingLogger.info("Unable to set file writable: "+file.getPath());
                }
            }
        }
    }

    /**
     * Changing file on os to read only and get back a new instance of the writable BIFile
     */
    public static BIFile getWritableBIFile(BIFile file) {
        if(!getFile(file).setWritable(true,false)){
            EasyTemplatingLogger.info("Unable to set file writable: "+file.getFilePath());
        }
        return (BIFile) file.getOrdInHost().resolve().get();
    }
    /**
     * Changing file on os to read only and get back a new instance of the writable BPxFile
     */
    public static BPxFile getWritableBPxFile(BIFile file) {
        if(!getFile(file).setWritable(true, false)){
            EasyTemplatingLogger.info("Unable to set file writable: "+file.getFilePath());
        }
        return (BPxFile) file.getOrdInHost().resolve().get();
    }

    /**
     * Function returning the station name either with fox/foxs connection
     */
    public static String getStationName() {
        try {
            return ((BStation) BOrd.make("local:|foxs:|station:|slot:/").resolve().get()).getStationName();
        } catch (Exception ex) {
            EasyTemplatingLogger.error(ex.getMessage(), ex);
            try {
                return ((BStation) BOrd.make("local:|fox:|station:|slot:/").resolve().get()).getStationName();
            } catch (Exception ex1) {
                EasyTemplatingLogger.error(ex1.getMessage(), ex1);
            }
        }
        return Sys.getStation().getStationName();
    }

    /**
     * Function returning the file path to the file from BIFile
     */
    private static String getFilePath(BIFile file) {
        String ord = file.getOrdInHost().toString();
        if (ord.contains("~")) {
            return ord.substring(ord.indexOf('~')
                    + 1, ord.length());
        }
        return ord.substring(ord.indexOf('^') + 1, ord.length());
    }

    /**
     * Returning a file from the core op system
     */
    public static File getFile(BIFile file) {
        StringBuilder sPath = new StringBuilder();
        if (file.getOrdInHost().toString().contains("~")) {
            // Niagara Home file
            sPath.append(Sys.getNiagaraUserHome()).append(File.separator)
                    .append(getFilePath(file));
        } else {
            //station file
            sPath.append(Sys.getNiagaraUserHome()).append(File.separator)
                    .append("stations").append(File.separator).append(getStationName())
                    .append(File.separator).append("shared").append(File.separator)
                    .append(getFilePath(file));
        }

        return Paths.get(sPath.toString()).toFile();
    }


    public static BOrd getCommonRootForComplexObject(boolean isSimpleObjectCreation,
                                               boolean isWorkbench, BIFile... selectedFile)
            throws EasyTemplatingBuilderException {

        // Init global variables

        pxFiles.clear();
        popupPxFiles.clear();

        // Get all bindings from all px files

        List<BOrd> allBindings = getAllBindingsFromObjectAndPopupFiles(false, isWorkbench, selectedFile);

        // Removing any relative ords

        List<BOrd> notRelativised = new ArrayList<>();
        for (BOrd ord : allBindings) {
            boolean startsWithSlot = isRelativised(ord.toString());
            if (!startsWithSlot) {
                notRelativised.add(ord);
            }
        }
        allBindings = notRelativised;

        if (allBindings.isEmpty()) {
            return BOrd.NULL;
        }

        // Now absoluteOrds will be never empty

        BOrd currentParent = allBindings.get(0);
        if (isControlPoint(currentParent)) {
            currentParent = currentParent.getParent();
        }

        while (currentParent != null) {
            if (!bindingsAllContainParent(allBindings, currentParent)) {
                currentParent = currentParent.getParent();
            } else {
                break;
            }
        }

        if (currentParent != null) {
            if (currentParent.toString().equals(BASE_SLOT)) {
                return BOrd.make(BASE_SLOT);
            } else {
                //add all files from root -> subFolders (views and popup.px files)
                addRootSubFoldersViewsAndPopupFiles(isSimpleObjectCreation, currentParent);
                return currentParent;
            }
        } else {
            return BOrd.make(BASE_SLOT);
        }

    }

    private static boolean isControlPoint(BOrd ord) {
        return ord.get(Sys.getStation()) instanceof BControlPoint;
    }

    private static List<BOrd> getAllEtsoBindings(BIFile selectedFile){
        List<BOrd> output = new ArrayList<>();
        try{
            PxDecoder decodedPx = new PxDecoder(selectedFile);
            BWidget rootWidget = decodedPx.decodeDocument();
            List<BBinding> bindings = getBindingsRecursive(rootWidget);
            Type easyTemplatingBindingType = Sys.getType(ET_EASY_TEMPLATING_BINDING_TYPE);
            for (BBinding binding : bindings) {
                if (binding.getType().is(easyTemplatingBindingType)) {
                    output.add(BOrd.make(WB_FILE_BASE_SLOT + ((BComponent)binding.get("data")).get("data").toString()));
                }
            }
        }catch (final Exception exp){
            EasyTemplatingLogger.infoWithException("Error: EasyTemplatingUtils: Error while getting all ETSO Bindings ", exp);
        }
        return output;
    }

    public static List<BPxInclude> getAllPxIncludesThatContainEasyTemplatingBinding(BIFile selectedFile) {
        List<BPxInclude> output = new ArrayList<>();
        try {
            PxDecoder decoderPx = new PxDecoder(selectedFile);
            BWidget rootWidget = decoderPx.decodeDocument();
            List<BBinding> bindings = getBindingsRecursive(rootWidget);
            Type easyTemplatingBindingType = Sys.getType(ET_EASY_TEMPLATING_BINDING_TYPE);
            for (BBinding binding : bindings) {
                if (binding.getType().is(easyTemplatingBindingType) &&
                    binding.getParent() instanceof BPxInclude) {
                    output.add((BPxInclude) binding.getParent());
                }
            }
        } catch (final Exception exp) {
            if (LOGGER.isLoggable(Level.FINER)){
                LOGGER.log(Level.FINER,"Failed to decode px file: " + selectedFile.getNavOrd().toString(), exp);
            }
        }
        return output;
    }

    public static List<BIFile> loadAllFilesToList(BDirectory fileDirectory){
        List<BIFile> output = loadAllEtcoFilesInAllFolders(fileDirectory);
        for(BIFile file : fileDirectory.listFiles()){
            if(!file.isDirectory() && file.getExtension().equals(ET_COMPLEX_OBJ_FILE_EXT)){
                output.add(file);
            }
        }
        return output;
    }

    public static List<BIFile> loadAllEtcoFilesInAllFolders(BDirectory fileDirectory){
        List<BIFile> output = new ArrayList<>();
        for(BINavNode node: fileDirectory.getNavChildren())
        {
            List<BIFile> tempList = checkFileByFile(node);
            output.addAll(tempList);
        }
        return output;
    }

    public static List<BIFile> checkFileByFile(BINavNode node){
        List<BIFile> output = new ArrayList<>();
        try {
            Iterator<BINavNode> nodeIterator = node.iterateNavDescendants();
            while (nodeIterator.hasNext()) {
                BINavNode childNode = nodeIterator.next();
                if (childNode instanceof BIFile
                        && !((BIFile)childNode).isDirectory()
                        && ((BIFile)childNode).getExtension().equals(ET_COMPLEX_OBJ_FILE_EXT)) {
                    output.add((BIFile)childNode);
                }
            }
        }catch (final Exception exp){
            if(LOGGER.isLoggable(Level.FINER)){
                LOGGER.log(Level.FINER,"Error while looking etco files under all the folders.",exp);
            }
        }
        return output;
    }

    public static void getInvalidOrdsFromPxFile(
            BIFile pxFile,
            String commonRoot,
            Set<String> invalidOrds,
            Set<String> invalidPopupBindingOrds) {

        try {
            PxDecoder decoder = new PxDecoder(pxFile);
            BWidget rootWidget = decoder.decodeDocument();
            List<BBinding> bindings = getBindingsRecursive(rootWidget);
            for (BBinding binding : bindings) {
                if (binding.getType().is(Sys.getType(ET_EASY_POPUP_BINDING_TYPE))) {
                    // Nothing to do
                    // If EasyPopupBinding appears, that means this px file is .etso
                    // And the popup binding will be handled when addressing .etco
                    // Because an EasyTemplatingBinding can have etsoOrd and its popupOrd
                } else if (binding.getType().is(Sys.getType(ET_EASY_TEMPLATING_BINDING_TYPE))) {
                    getInvalidOrdsFromEasyTemplatingBinding(binding, invalidOrds, invalidPopupBindingOrds);
                } else if (binding instanceof BPopupBinding) {
                    if (!binding.getOrd().toString().contains(ET_VIEW_REQ_EXT)) {
                        getInvalidOrdsFromPopupBinding(binding, invalidOrds, invalidPopupBindingOrds);
                    }
                } else {
                    getInvalidOrdsFromOtherBinding(binding, commonRoot, invalidOrds);
                }
            }
        }
        catch (Exception ex) {
            if (LOGGER.isLoggable(Level.FINER)){
                LOGGER.log(Level.FINER, "Cannot decode px file: " + pxFile.getNavOrd().toString(), ex);
            }
        }
    }

    private static void getInvalidOrdsFromEasyTemplatingBinding(
            BBinding binding,
            Set<String> invalidOrds,
            Set<String> invalidPopupBindingOrds) {

        getInvalidOrdsFromEasyTemplatingBindingEtso(binding, invalidOrds, invalidPopupBindingOrds);
        getInvalidOrdsFromEasyTemplatingBindingEtsoPopup(binding, invalidOrds, invalidPopupBindingOrds);
    }

    private static void getInvalidOrdsFromEasyTemplatingBindingEtso(
            BBinding binding,
            Set<String> invalidOrds,
            Set<String> invalidPopupBindingOrds) {

        BPxInclude pxInclude = (BPxInclude) binding.getParent();
        String commonRoot = pxInclude.getVariables().get("var").toString();
        BOrd etsoOrd = pxInclude.getOrd();

        try {
            BIFile etsoFile = (BIFile) BOrd.make(ET_WB_COMPONENT_SPACE + etsoOrd.toString()).resolve().get();
            getInvalidOrdsFromPxFile(etsoFile, commonRoot, invalidOrds, invalidPopupBindingOrds);
        } catch (Exception ex) {
            if (LOGGER.isLoggable(Level.FINER)){
                LOGGER.log(Level.FINER, "Cannot resolve px ord: " + etsoOrd.toString(), ex);
            }
            invalidOrds.add(etsoOrd.toString());
        }
    }

    private static void getInvalidOrdsFromEasyTemplatingBindingEtsoPopup(
            BBinding binding,
            Set<String> invalidOrds,
            Set<String> invalidPopupBindingOrds) {

        BPxInclude pxInclude = (BPxInclude) binding.getParent();
        String commonRoot = pxInclude.getVariables().get("var").toString();
        BOrd popupOrd = (BOrd) pxInclude.get("popupOrd");
        BOrd realPopupOrd = null;

        if (popupOrd == null || popupOrd.isNull()) {
            // Here means EasyTemplatingBinding has no popup
            return;
        }
        String popupOrdStr = popupOrd.toString().replace(ET_WB_COMPONENT_SPACE, "");

        try {
            BIFile popupFile = (BIFile) popupOrd.resolve().get();
            PxDecoder decoder = new PxDecoder(popupFile);
            BWidget rootWidget = decoder.decodeDocument();
            BPxInclude pxIncludeWidget = getPxIncludeOfWidget(rootWidget);
            if (pxIncludeWidget == null) {
                invalidPopupBindingOrds.add(popupOrdStr);
                return;
            } else {
                realPopupOrd = pxIncludeWidget.getOrd();
            }
        } catch (Exception ex) {
            if (LOGGER.isLoggable(Level.FINER)){
                LOGGER.log(Level.FINER, "Cannot resolve px ord: " + popupOrdStr, ex);
            }
            invalidPopupBindingOrds.add(popupOrdStr);
            return;
        }

        String realPopupOrdStr = realPopupOrd.toString();
        try {
            BIFile realPopupFile = (BIFile) BOrd.make(ET_WB_COMPONENT_SPACE + realPopupOrdStr).resolve().get();
            getInvalidOrdsFromPxFile(realPopupFile, commonRoot, invalidOrds, invalidPopupBindingOrds);
        } catch (Exception ex) {
            if (LOGGER.isLoggable(Level.FINER)){
                LOGGER.log(Level.FINER, "Cannot resolve px ord: " + realPopupOrdStr, ex);
            }
            invalidPopupBindingOrds.add(realPopupOrdStr);
        }
    }

    private static BPxInclude getPxIncludeOfWidget(BWidget widget) {
        if (widget instanceof BPxInclude) {
            return (BPxInclude) widget;
        }
        BWidget[] childs = widget.getChildWidgets();
        for (BWidget child : childs) {
            if (child instanceof BPxInclude) {
                return (BPxInclude) child;
            } else if (child.getChildWidgets().length != 0) {
                return getPxIncludeOfWidget(child);
            }
        }
        return null;
    }

    private static void getInvalidOrdsFromPopupBinding(
            BBinding binding,
            Set<String> invalidOrds,
            Set<String> invalidPopupBindingOrds) {

        String ordStr = binding.getOrd().toString();
        if (isEmptyOrd(ordStr)) {
            invalidPopupBindingOrds.add(ordStr);
            return;
        }
        
        try {
            if (ordStr.startsWith(ET_FILE_OS)) {
                BOrd popupOrd = BOrd.make(ET_WB_LOCAL_SPACE + ordStr);
                BIFile popupFile = (BIFile) popupOrd.resolve().get();
                getInvalidOrdsFromPxFile(popupFile, null, invalidOrds, invalidPopupBindingOrds);
            } else if (ordStr.startsWith(ET_FILE)) {
                BOrd popupOrd = BOrd.make(ET_WB_COMPONENT_SPACE + ordStr);
                BIFile popupFile = (BIFile) popupOrd.resolve().get();
                getInvalidOrdsFromPxFile(popupFile, null, invalidOrds, invalidPopupBindingOrds);
            } else {
                BOrd popupOrd = BOrd.make(ET_WB_COMPONENT_SPACE + ordStr);
                popupOrd.resolve().get();
            }
        } catch (Exception ex) {
            if (LOGGER.isLoggable(Level.FINER)){
                LOGGER.log(Level.FINER, "Ord cannot be resolved: " + ordStr, ex);
            }
            invalidPopupBindingOrds.add(ordStr);
        }
    }

    private static void getInvalidOrdsFromOtherBinding(
             BBinding binding,
             String commonRoot,
             Set<String> invalidOrds) {

        if (isNeedlessBindingInstanceOfEasyBindingWidget(binding)) {
            return;
        }
        String widgetName= binding.getWidget().getName();
        Map<String, BOrd> ordInfo = EasyTemplatingUtils.getRealOrd(binding);
        for (Map.Entry<String, BOrd> entrySet : ordInfo.entrySet()) {

            String slotName = entrySet.getKey();
            BOrd ord = entrySet.getValue();
            String ordStr = ord.toString();
            if (slotName.equals("hyperlink") && ord.isNull()) {
                continue;
            }

            if (isEmptyOrd(ordStr)) {
                if(!binding.getType().getTypeName().equals(ET_EASY_HISTORY_POPUP_BINDING)) {
                    invalidOrds.add(lex.getText("EasyTemplate.warning.dialog.empty.ord", new Object[]{binding.getType(), widgetName}));
                }
                continue;
            }
            if (commonRoot != null) {
                ordStr = ordStr.replace(ET_BINDING_ORD_PREFIX, commonRoot);
            }
            if (isRelativised(ordStr)) {
                continue;
            }

            try {
                BOrd.make(ET_WB_COMPONENT_SPACE + ordStr).resolve().get();
            } catch (Exception ex) {
                if (LOGGER.isLoggable(Level.FINER)){
                    LOGGER.log(Level.FINER, "Ord cannot be resolved: " + ordStr, ex);
                }
                invalidOrds.add(ordStr+lex.getText("EasyTemplate.warning.dialog.invalid.slot", new Object[]{binding.getType(),widgetName}));
            }
        }
    }

    public static boolean isHonBacnetVirtualPropertyBinding(BBinding binding) {
        String bindingType = binding.getType().toString();
        return bindingType.equals(ET_HON_BACNET_VIRTUAL_PROPERTY_BINDING);
    }

    public static boolean isHonBacnetHistoryBinding(BBinding binding) {
        String bindingType = binding.getType().toString();
        return bindingType.equals(ET_HON_BACNET_HISTORY_BINDING);
    }

    public static Map<String, BOrd> getRealOrd(BBinding binding) {
        Map<String, BOrd> ordInfo = new HashMap<>();
        if (isHonBacnetVirtualPropertyBinding(binding)) {
            ordInfo.put("sourceOrdClone", (BOrd) binding.get("sourceOrdClone"));
        } else if (isHonBacnetHistoryBinding(binding)) {
            ordInfo.put("originatingOrd", (BOrd) binding.get("originatingOrd"));
        } else if (binding instanceof BValueBinding) {
            BOrd hyperlinkOrd = (BOrd) binding.get("hyperlink");
            if (hyperlinkOrd != null && !hyperlinkOrd.isNull()) {
                ordInfo.put("hyperlink", hyperlinkOrd);
            }
            ordInfo.put("ord", binding.getOrd());
        } else {
            ordInfo.put("ord", binding.getOrd());
        }
        return ordInfo;
    }

    public static void setRealOrd(BBinding binding, String slotName, BOrd ord) {
        binding.set(slotName, ord);
    }

    private static boolean isNeedlessBindingInstanceOfEasyBindingWidget(BBinding binding) {
        String bindingType = binding.getType().toString();
        String widgetType = binding.getWidget().getType().toString();
        return (bindingType.equals(ET_EASY_VALUE_BINDING_TYPE) ||
            bindingType.equals(ET_EASY_ALARM_BINDING_TYPE) ||
            bindingType.equals(ET_EASY_OVERRIDE_BINDING_TYPE)) &&
            !widgetType.equals(ET_EASY_BINDING_WIDGET_TYPE);
    }

    private static boolean isHistoryPopupBinding(BBinding binding) {
        return binding.getOrd().toString().contains(ET_HISTORY_OS);
    }

    public static boolean isEmptyOrd(String ordStr) {
        return ordStr == null || ordStr.isEmpty() || ordStr.equals("null");
    }

    public static boolean isValidControlPoint(BOrd ord) {
        try {
            return ord.get(Sys.getStation()) instanceof BControlPoint;
        } catch (Exception ex) {
            if (LOGGER.isLoggable(Level.FINER)){
                LOGGER.log(Level.FINER, "Ord cannot be resolved: " + ord.toString(), ex);
            }
            return false;
        }
    }
    public static BIFile[] getTemlateFileAndDirectory(BPxInclude baseWidget){
        BIFile templateFile;
        BOrd etsoOrd = baseWidget.getOrd();
        String ord = etsoOrd.toString();
        OrdQuery[] queries = etsoOrd.parse();
        FilePath fp = (FilePath) queries[queries.length - 1];
        BSpace space;
        if (etsoOrd.toString().contains(ET_WB_LOCAL_FOXS)) {
            space = etsoOrd.resolve().getSpace();
        } else {
            space = BOrd.make(ord.contains(ET_WB_LOCAL_FOXS ) ? ord: ET_WB_LOCAL_FOXS.concat(ord)).resolve().getSpace();
        }
        if (space.getType().equals(BFileSystem.TYPE)) {
            templateFile = BFileSystem.INSTANCE.findFile(fp);
        } else {
            templateFile = ((BFoxFileSpace) space).findFile(fp);
        }
        String fileToCopyPath = templateFile.getFilePath().toString();
        String etsoDirectory = fileToCopyPath.substring(0, fileToCopyPath.lastIndexOf('/'));
        BDirectory directory;
        directory = (BDirectory) BOrd.make(etsoDirectory).get(space);


        return new BIFile[]{templateFile,directory};
    }
    private static List<BOrd> pxFiles = new ArrayList<>();
    private static List<PxFileRef> popupPxFiles = new ArrayList<>();

    public static Set<String> readXml(BIFile file) {
        Set<String> ordList = new HashSet<>();
        try {
            DocumentBuilderFactory builderFactory = newInstance();
            DocumentBuilder documentBuilder = builderFactory.newDocumentBuilder();
            Document document = documentBuilder.parse(file.getInputStream());
            //If Px includes with ValueBinding
            NodeList easyValueBindingOrdList = document.getElementsByTagName(VALUEBINDING);
            for (int i = 0; i < easyValueBindingOrdList.getLength(); i++) {
                Node node = easyValueBindingOrdList.item(i);
                addValueBindingOrdEntries(node, ordList);
            }
            //If Px includes with BoundLabel
            NodeList boundLabelOrdList = document.getElementsByTagName(BOUNDLABEL);
            for (int i = 0; i < boundLabelOrdList.getLength(); i++) {
                Node node = boundLabelOrdList.item(i);
                NodeList childList = node.getChildNodes();
                addBoundLabelOrdEntries(childList, ordList);
            }
            //If Px includes with EB EasyBindingWidget
            NodeList ebOrdList = document.getElementsByTagName(EB_EASYBINDINGWIDGET);
            for (int i = 0; i < ebOrdList.getLength(); i++) {
                Node node = ebOrdList.item(i);
                NodeList childList = node.getChildNodes();
                addEBWidgetOrdEntries(childList, ordList);
            }
        } catch (Exception ex) {
            if (LOGGER.isLoggable(Level.FINER)){
                LOGGER.log(Level.FINER, "error in read xml for et file ", ex);
            }
        }
        return ordList;
    }

    private static void addEBWidgetOrdEntries(NodeList childList, Set<String> ordList)
    {
        ArrayList<String> ebOrdEntries = new ArrayList<String>();
        for (int j = 0; j < childList.getLength(); j++) {
            if (childList.item(j).getNodeType() == Node.ELEMENT_NODE) {
                if(childList.item(j).getAttributes().item(1) != null) {
                    if (childList.item(j).getAttributes().item(1).toString().contains("ord=\"$(var)")){
                        ebOrdEntries.add(childList.item(j).getAttributes().item(1).toString());
                    }
                }
            }
        }
        HashSet<String> ebOrdUniqueEntries = new HashSet(ebOrdEntries);
        for(String ebOrdUniqueEntry : ebOrdUniqueEntries) {
            ordList.add(ebOrdUniqueEntry);
        }
    }

    private static void addBoundLabelOrdEntries(NodeList childList, Set<String> ordList)
    {
        for (int j = 0; j < childList.getLength(); j++) {
            if (childList.item(j).getNodeType() == Node.ELEMENT_NODE) {
                NamedNodeMap map = childList.item(j).getAttributes();
                for (int i = 0; i < map.getLength(); i++) {
                    if (map.item(i) != null && map.item(i).toString().contains("ord=\"$(var)")) {
                        ordList.add(map.item(i).toString());
                    }
                    if (map.item(i) != null && map.item(i).toString().contains("hyperlink=\"$(var)")) {
                        String hyperlinkStr = map.item(i).toString();
                        hyperlinkStr = hyperlinkStr.replace("hyperlink=", "ord=");
                        int viewIndex = hyperlinkStr.indexOf("|view:");
                        if (viewIndex != -1) {
                            hyperlinkStr = hyperlinkStr.substring(0, viewIndex);
                            hyperlinkStr += "\"";
                        }
                        ordList.add(hyperlinkStr);
                    }
                }
            }
        }
    }

    private static void addValueBindingOrdEntries(Node node, Set<String> ordList) {
        ArrayList<String> ebOrdEntries = new ArrayList<String>();
        if ((node).getAttributes().item(0) != null) {
            if ((node).getAttributes().item(0).toString().contains("ord=\"$(var)")) {
                ebOrdEntries.add((node).getAttributes().item(0).toString());
            }
        }
        HashSet<String> ebOrdUniqueEntries = new HashSet(ebOrdEntries);
        for (String ebOrdUniqueEntry : ebOrdUniqueEntries) {
            ordList.add(ebOrdUniqueEntry);
        }
    }

}